import itertools
import re
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Collection, Iterable, Sequence
from dataclasses import dataclass, field
from graphlib import CycleError, TopologicalSorter
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, TypeVar, cast, overload

from cognite.client.data_classes import filters
from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling import (
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerList,
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelList,
    EdgeConnection,
    MappedProperty,
    Node,
    NodeApply,
    NodeApplyList,
    NodeList,
    RequiresConstraint,
    Space,
    SpaceApply,
    SpaceApplyList,
    SpaceList,
    View,
    ViewApply,
    ViewApplyList,
    ViewList,
)
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    DataModelId,
    NodeId,
    ViewId,
)
from cognite.client.data_classes.data_modeling.views import (
    EdgeConnectionApply,
    MappedPropertyApply,
    ReverseDirectRelation,
    ReverseDirectRelationApply,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite.neat._v0.core._client.data_classes.data_modeling import Component
from cognite.neat._v0.core._client.data_classes.schema import DMSSchema
from cognite.neat._v0.core._issues.warnings import CDFMaxIterationsWarning
from cognite.neat._v0.core._shared import T_ID
from cognite.neat._v0.core._utils.tarjan import tarjan

if TYPE_CHECKING:
    from cognite.neat._v0.core._client._api_client import NeatClient

T_WritableCogniteResourceList = TypeVar("T_WritableCogniteResourceList", bound=WriteableCogniteResourceList)

T_Item = TypeVar("T_Item")
T_Out = TypeVar("T_Out", bound=Iterable)


@dataclass
class MultiCogniteAPIError(Exception, Generic[T_ID, T_WritableCogniteResourceList]):
    success: T_WritableCogniteResourceList
    failed: list[T_ID] = field(default_factory=list)
    errors: list[CogniteAPIError] = field(default_factory=list)


class ResourceLoader(
    ABC,
    Generic[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList],
):
    """A resource loaders is a wrapper around the Cognite SDK that does a few things:

    * Standardizes the CRUD operations for a given resource.
    * Caches the items that have been retrieved from the CDF.
    """

    resource_name: str
    dependencies: "ClassVar[frozenset[type[ResourceLoader]]]" = frozenset()

    def __init__(self, client: "NeatClient") -> None:
        # This is exposed to allow for disabling the cache.
        self.cache = True
        self._client = client
        # This cache is used to store the items that have been retrieved from the CDF.
        self._items_by_id: dict[T_ID, T_WritableCogniteResource] = {}

    def bust_cache(self) -> None:
        self._items_by_id = {}

    @classmethod
    @abstractmethod
    def get_id(cls, item: T_WriteClass | T_WritableCogniteResource | dict | T_ID) -> T_ID:
        raise NotImplementedError

    @classmethod
    def get_ids(cls, items: Sequence[T_WriteClass | T_WritableCogniteResource]) -> list[T_ID]:
        return [cls.get_id(item) for item in items]

    def create(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        # Containers can have dependencies on other containers, so we sort them before creating them.
        items = self.sort_by_dependencies(items)

        exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
        try:
            created = self._fallback_one_by_one(self._create, items)
        except MultiCogniteAPIError as e:
            created = e.success
            exception = e

        if self.cache:
            self._items_by_id.update({self.get_id(item): item for item in created})

        if exception is not None:
            raise exception

        return created

    @overload
    def retrieve(
        self, ids: SequenceNotStr[T_ID], format: Literal["read"] = "read"
    ) -> T_WritableCogniteResourceList: ...

    @overload
    def retrieve(self, ids: SequenceNotStr[T_ID], format: Literal["write"] = "write") -> T_CogniteResourceList: ...

    def retrieve(
        self, ids: SequenceNotStr[T_ID], format: Literal["read", "write"] = "read"
    ) -> T_WritableCogniteResourceList | T_CogniteResourceList:
        if not self.cache:
            # We now that SequenceNotStr = Sequence
            output = self._fallback_one_by_one(self._retrieve, ids)  # type: ignore[arg-type]
        else:
            exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
            missing_ids = [id for id in ids if id not in self._items_by_id.keys()]
            if missing_ids:
                try:
                    retrieved = self._retrieve(missing_ids)
                except MultiCogniteAPIError as e:
                    retrieved = e.success
                    exception = e
                self._items_by_id.update({self.get_id(item): item for item in retrieved})
            if exception is not None:
                raise exception
            # We need to check the cache again, in case we didn't retrieve all the items.
            output = self._create_list([self._items_by_id[id] for id in ids if id in self._items_by_id])
        if format == "write":
            return cast(T_CogniteResourceList, output.as_write())
        return output

    def update(
        self, items: Sequence[T_WriteClass], force: bool = False, drop_data: bool = False
    ) -> T_WritableCogniteResourceList:
        exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
        if force:
            updated = self._update_force(items, drop_data=drop_data)
        else:
            try:
                updated = self._fallback_one_by_one(self._update, items)
            except MultiCogniteAPIError as e:
                updated = e.success
                exception = e

        if self.cache:
            self._items_by_id.update({self.get_id(item): item for item in updated})

        if exception is not None:
            raise exception

        return updated

    def delete(self, ids: SequenceNotStr[T_ID] | Sequence[T_WriteClass]) -> list[T_ID]:
        id_list = [self.get_id(item) for item in ids]
        exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
        try:
            # We know that SequenceNotStr = Sequence
            deleted = self._fallback_one_by_one(self._delete, id_list)  # type: ignore[arg-type]
        except MultiCogniteAPIError as e:
            deleted = e.success
            exception = e

        if self.cache:
            for id in deleted:
                self._items_by_id.pop(id, None)
        if exception is not None:
            raise exception

        return deleted

    @abstractmethod
    def _create(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def _retrieve(self, ids: SequenceNotStr[T_ID]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def _update(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def _delete(self, ids: SequenceNotStr[T_ID]) -> list[T_ID]:
        raise NotImplementedError

    @abstractmethod
    def _create_list(self, items: Sequence[T_WritableCogniteResource]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    def has_data(self, item_id: T_ID) -> bool:
        return False

    def are_equal(self, local: T_WriteClass, remote: T_WritableCogniteResource) -> bool:
        return local == remote.as_write()

    def sort_by_dependencies(self, items: Sequence[T_WriteClass]) -> list[T_WriteClass]:
        return list(items)

    def _update_force(
        self,
        items: Sequence[T_WriteClass],
        drop_data: bool = False,
        tried_force_update: set[T_ID] | None = None,
        success: T_WritableCogniteResourceList | None = None,
    ) -> T_WritableCogniteResourceList:
        tried_force_update = tried_force_update or set()
        try:
            return self._update(items)
        except CogniteAPIError as e:
            failed_ids = {self.get_id(failed) for failed in itertools.chain(e.failed, e.unknown)}
            success_ids = [self.get_id(success) for success in e.successful]
            success_ = self.retrieve(success_ids)
            if success is None:
                success = success_
            else:
                success.extend(success_)
            to_redeploy: list[T_WriteClass] = []
            for item in items:
                item_id = self.get_id(item)
                if item_id in failed_ids:
                    if tried_force_update and item_id in tried_force_update:
                        # Avoid infinite loop
                        continue
                    tried_force_update.add(item_id)
                    if self.has_data(item_id) and not drop_data:
                        continue
                    to_redeploy.append(item)
            if not to_redeploy:
                # Avoid infinite loop
                raise e
            self.delete(to_redeploy)
            forced = self._update_force(to_redeploy, drop_data, tried_force_update, success)
            forced.extend(success)
            return forced

    def _fallback_one_by_one(self, method: Callable[[Sequence[T_Item]], T_Out], items: Sequence[T_Item]) -> T_Out:
        try:
            return method(items)
        except CogniteAPIError as e:
            exception = MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList](self._create_list([]))
            success = {self.get_id(success) for success in e.successful}
            if success:
                # Need read version of the items to put into cache.
                retrieve_items = self.retrieve(list(success))
                exception.success.extend(retrieve_items)
            for item in items:
                # We know that item is either T_ID or T_WriteClass
                # but the T_Item cannot be bound to both types at the same time.
                item_id = self.get_id(item)  # type: ignore[arg-type]
                if item_id in success:
                    continue
                try:
                    item_result = method([item])
                except CogniteAPIError as item_exception:
                    exception.errors.append(item_exception)
                    exception.failed.extend(self.get_ids(item_exception.failed))
                else:
                    exception.success.extend(item_result)
            raise exception from None


class DataModelingLoader(
    ResourceLoader[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList],
    ABC,
):
    support_merge: ClassVar[bool] = True

    @classmethod
    def in_space(cls, item: T_WriteClass | T_WritableCogniteResource | T_ID, space: set[str]) -> bool:
        if hasattr(item, "space"):
            return item.space in space
        raise ValueError(f"Item {item} does not have a space attribute")

    @classmethod
    @abstractmethod
    def items_from_schema(cls, schema: DMSSchema) -> T_CogniteResourceList:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def merge(cls, local: T_WriteClass, remote: T_WritableCogniteResource) -> T_WriteClass:
        raise NotImplementedError


class SpaceLoader(DataModelingLoader[str, SpaceApply, Space, SpaceApplyList, SpaceList]):
    resource_name = "spaces"
    support_merge = False

    @classmethod
    def get_id(cls, item: Space | SpaceApply | str | dict) -> str:
        if isinstance(item, Space | SpaceApply):
            return item.space
        if isinstance(item, dict):
            return item["space"]
        return item

    @classmethod
    def merge(cls, local: SpaceApply, remote: Space) -> SpaceApply:
        raise NotImplementedError("Spaces cannot be merged")

    def _create(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self._client.data_modeling.spaces.apply(items)

    def _retrieve(self, ids: SequenceNotStr[str]) -> SpaceList:
        return self._client.data_modeling.spaces.retrieve(ids)

    def _update(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[str] | Sequence[Space | SpaceApply]) -> list[str]:
        if all(isinstance(item, Space) for item in ids) or all(isinstance(item, SpaceApply) for item in ids):
            ids = [cast(Space | SpaceApply, item).space for item in ids]
        return self._client.data_modeling.spaces.delete(cast(SequenceNotStr[str], ids))

    def _create_list(self, items: Sequence[Space]) -> SpaceList:
        return SpaceList(items)

    def clean(self, space: str) -> None:
        """Deletes all data in a space.

        This means all nodes, edges, views, containers, and data models located in the given space.

        Args:
            client: Connected CogniteClient
            space: The space to delete.

        """
        edges = self._client.data_modeling.instances.list(
            "edge", limit=-1, filter=filters.Equals(["edge", "space"], space)
        )
        if edges:
            instances = self._client.data_modeling.instances.delete(edges=edges.as_ids())
            print(f"Deleted {len(instances.edges)} edges")
        nodes = self._client.data_modeling.instances.list(
            "node", limit=-1, filter=filters.Equals(["node", "space"], space)
        )
        node_types = {NodeId(node.type.space, node.type.external_id) for node in nodes if node.type}
        node_data = set(nodes.as_ids()) - node_types
        if node_data:
            instances = self._client.data_modeling.instances.delete(nodes=list(node_data))
            print(f"Deleted {len(instances.nodes)} nodes")
        if node_types:
            instances = self._client.data_modeling.instances.delete(nodes=list(node_types))
            print(f"Deleted {len(instances.nodes)} node types")
        views = self._client.data_modeling.views.list(limit=-1, space=space)
        if views:
            deleted_views = self._client.data_modeling.views.delete(views.as_ids())
            print(f"Deleted {len(deleted_views)} views")
        containers = self._client.data_modeling.containers.list(limit=-1, space=space)
        if containers:
            deleted_containers = self._client.data_modeling.containers.delete(containers.as_ids())
            print(f"Deleted {len(deleted_containers)} containers")
        if data_models := self._client.data_modeling.data_models.list(limit=-1, space=space):
            deleted_data_models = self._client.data_modeling.data_models.delete(data_models.as_ids())
            print(f"Deleted {len(deleted_data_models)} data models")
        deleted_space = self._client.data_modeling.spaces.delete(space)
        print(f"Deleted space {deleted_space}")

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> SpaceApplyList:
        return SpaceApplyList(schema.spaces.values())

    def has_data(self, item_id: str) -> bool:
        if self._client.data_modeling.instances.list("node", limit=1, space=item_id):
            return True
        if self._client.data_modeling.instances.list("edge", limit=1, space=item_id):
            return True
        # Need to check if there are any containers with data in the space. Typically,
        # a schema space will not contain data, while it will have containers that have data in an instance space.
        for container in self._client.data_modeling.containers(space=item_id, include_global=False):
            if self._client.loaders.containers.has_data(container.as_id()):
                return True
        return False


class ContainerLoader(DataModelingLoader[ContainerId, ContainerApply, Container, ContainerApplyList, ContainerList]):
    resource_name = "containers"
    dependencies = frozenset({SpaceLoader})

    @classmethod
    def get_id(cls, item: Container | ContainerApply | ContainerId | dict) -> ContainerId:
        if isinstance(item, Container | ContainerApply):
            return item.as_id()
        if isinstance(item, dict):
            return ContainerId.load(item)
        return item

    @classmethod
    def merge(cls, local: ContainerApply, remote: Container) -> ContainerApply:
        if local.as_id() != remote.as_id():
            raise ValueError(f"Cannot merge containers with different IDs: {local.as_id()} and {remote.as_id()}")
        if local.used_for != remote.used_for:
            raise ValueError(f"Cannot merge containers with different used_for: {local.used_for} and {remote.used_for}")
        remote_write = remote.as_write()
        existing_properties = remote_write.properties or {}
        merged_properties = {**existing_properties, **(local.properties or {})}
        merged_indices = {**remote_write.indexes, **local.indexes}
        merged_constraints = {**remote_write.constraints, **local.constraints}
        return ContainerApply(
            space=remote.space,
            external_id=remote.external_id,
            properties=merged_properties,
            description=local.description or remote.description,
            name=local.name or remote.name,
            used_for=local.used_for,
            constraints=merged_constraints,
            indexes=merged_indices,
        )

    def sort_by_dependencies(self, items: Sequence[ContainerApply]) -> list[ContainerApply]:
        container_by_id = {container.as_id(): container for container in items}
        container_dependencies = {
            container.as_id(): {
                const.require
                for const in container.constraints.values()
                if isinstance(const, RequiresConstraint) and const.require in container_by_id
            }
            for container in items
        }
        return [
            container_by_id[container_id] for container_id in TopologicalSorter(container_dependencies).static_order()
        ]

    def _create(self, items: Sequence[ContainerApply]) -> ContainerList:
        return self._client.data_modeling.containers.apply(items)

    @overload
    def retrieve(
        self, ids: SequenceNotStr[ContainerId], format: Literal["read"] = "read", include_connected: bool = False
    ) -> ContainerList: ...

    @overload
    def retrieve(
        self, ids: SequenceNotStr[ContainerId], format: Literal["write"] = "write", include_connected: bool = False
    ) -> ContainerApplyList: ...

    def retrieve(
        self,
        ids: SequenceNotStr[ContainerId],
        format: Literal["read", "write"] = "read",
        include_connected: bool = False,
    ) -> ContainerList | ContainerApplyList:
        if not include_connected:
            return super().retrieve(ids)

        # Retrieve recursively updates the cache.
        output = self._retrieve_recursive(ids)
        if format == "write":
            return output.as_write()
        return output

    def _retrieve(self, ids: SequenceNotStr[ContainerId]) -> ContainerList:
        return self._client.data_modeling.containers.retrieve(cast(Sequence, ids))

    def _update(self, items: Sequence[ContainerApply]) -> ContainerList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[ContainerId]) -> list[ContainerId]:
        return self._client.data_modeling.containers.delete(cast(Sequence, ids))

    def _create_list(self, items: Sequence[Container]) -> ContainerList:
        return ContainerList(items)

    def _retrieve_recursive(self, container_ids: SequenceNotStr[ContainerId]) -> ContainerList:
        """Containers can reference each other through the 'requires' constraint.

        This method retrieves all containers that are referenced by other containers through the 'requires' constraint,
        including their parents.
        """
        max_iterations = 10  # Limiting the number of iterations to avoid infinite loops
        found = ContainerList([])
        found_ids: set[ContainerId] = set()
        last_batch = list(container_ids)
        for _ in range(max_iterations):
            if not last_batch:
                break
            to_retrieve_from_cdf: set[ContainerId] = set()
            batch_ids: list[ContainerId] = []
            for container_id in last_batch:
                if container_id in found_ids:
                    continue
                elif container_id in self._items_by_id:
                    container = self._items_by_id[container_id]
                    found.append(container)
                    batch_ids.extend(self.get_connected_containers(container, found_ids))
                else:
                    to_retrieve_from_cdf.add(container_id)

            if to_retrieve_from_cdf:
                retrieved_batch = self._client.data_modeling.containers.retrieve(list(to_retrieve_from_cdf))
                self._items_by_id.update({view.as_id(): view for view in retrieved_batch})
                found.extend(retrieved_batch)
                found_ids.update({view.as_id() for view in retrieved_batch})
                for container in retrieved_batch:
                    batch_ids.extend(self.get_connected_containers(container, found_ids))

            last_batch = batch_ids
        else:
            warnings.warn(
                CDFMaxIterationsWarning(
                    "The maximum number of iterations was reached while resolving referenced containers."
                    "There might be referenced containers that are not included in the list of containers.",
                    max_iterations=max_iterations,
                ),
                stacklevel=2,
            )

        if self.cache is False:
            # We must update the cache to retrieve recursively.
            # If the cache is disabled, bust the cache to avoid storing the retrieved views.
            self.bust_cache()
        return found

    @staticmethod
    def get_connected_containers(
        container: Container | ContainerApply, skip: set[ContainerId] | None = None
    ) -> set[ContainerId]:
        connected_containers = set()
        for constraint in container.constraints.values():
            if isinstance(constraint, RequiresConstraint):
                connected_containers.add(constraint.require)
        if skip:
            return {container_id for container_id in connected_containers if container_id not in skip}
        return connected_containers

    def are_equal(self, local: ContainerApply, remote: Container) -> bool:
        local_dumped = local.dump(camel_case=True)
        if "usedFor" not in local_dumped:
            # Setting used_for to "node" as it is the default value in the CDF.
            local_dumped["usedFor"] = "node"

        return local_dumped == remote.as_write().dump(camel_case=True)

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> ContainerApplyList:
        return ContainerApplyList(schema.containers.values())

    def has_data(self, item_id: ContainerId) -> bool:
        has_data_filter = filters.HasData(containers=[item_id])
        has_data = False
        instance_type: Literal["node", "edge"]
        # Mypy does not understand that the instance type is Literal["node", "edge"]
        for instance_type in ["node", "edge"]:  # type: ignore[assignment]
            try:
                has_data = bool(
                    self._client.data_modeling.instances.list(instance_type, limit=1, filter=has_data_filter)
                )
            except CogniteAPIError as e:
                if e.code != 400:
                    # If the container is used for nodes and we ask for edges, we get a 400 error. This
                    # means there is no edge data for this container.
                    raise
            if has_data:
                return True
        return has_data


class ViewLoader(DataModelingLoader[ViewId, ViewApply, View, ViewApplyList, ViewList]):
    resource_name = "views"
    dependencies = frozenset({SpaceLoader, ContainerLoader})

    @classmethod
    def get_id(cls, item: View | ViewApply | ViewId | dict) -> ViewId:
        if isinstance(item, View | ViewApply):
            return item.as_id()
        if isinstance(item, dict):
            return ViewId.load(item)
        return item

    @classmethod
    def merge(cls, local: ViewApply, remote: View) -> ViewApply:
        if local.as_id() != remote.as_id():
            raise ValueError(f"Cannot merge views with different IDs: {local.as_id()} and {remote.as_id()}")
        remote_write = remote.as_write()
        existing_properties = remote_write.properties or {}
        merged_properties = {**existing_properties, **(local.properties or {})}
        merged_implements = list(remote_write.implements or [])
        for view_id in local.implements or []:
            if view_id not in merged_implements:
                merged_implements.append(view_id)
        return ViewApply(
            space=remote.space,
            external_id=remote.external_id,
            version=remote.version,
            properties=merged_properties,
            description=local.description or remote.description,
            name=local.name or remote.name,
            filter=local.filter or remote.filter,
            implements=merged_implements,
        )

    def _create(self, items: Sequence[ViewApply]) -> ViewList:
        return self._client.data_modeling.views.apply(items)

    def create(self, items: Sequence[ViewApply]) -> ViewList:
        try:
            return self._create(items)
        except CogniteAPIError as e1:
            if self._is_auto_retryable(e1):
                # Fallback to creating one by one if the error is auto-retryable.
                return self._fallback_one_by_one(self._create, items)
            elif self._is_false_not_exists(e1, {item.as_id() for item in items}):
                return self._try_to_recover_coupled(items, e1)
            raise

    def _try_to_recover_coupled(self, items: Sequence[ViewApply], original_error: CogniteAPIError) -> ViewList:
        """The /models/views endpoint can give faulty 400 about missing views that are part of the request.

        This method tries to recover from such errors by identifying the strongly connected components in the graph
        defined by the implements and through properties of the views. We then create the components in topological
        order.

        Args:
            items: The items that failed to create.
            original_error: The original error that was raised. If the problem is not recoverable, this error is raised.

        Returns:
            The views that were created.

        """
        views_by_id = {self.get_id(item): item for item in items}
        parents_by_child: dict[ViewId, set[ViewId]] = {
            view_id: {parent for parent in view.implements or [] if parent in views_by_id}
            for view_id, view in views_by_id.items()
        }
        # Check for cycles in the implements graph
        try:
            TopologicalSorter(parents_by_child).static_order()
        except CycleError as e:
            raise CycleError(f"Failed to deploy views. This likely due to a cycle in implements. {e.args[1]}") from None

        dependencies_by_id: dict[ViewId, set[ViewId]] = defaultdict(set)
        for view_id, view in views_by_id.items():
            dependencies_by_id[view_id].update([parent for parent in view.implements or [] if parent in views_by_id])
            for properties in (view.properties or {}).values():
                if isinstance(properties, ReverseDirectRelationApply):
                    if isinstance(properties.through.source, ViewId) and properties.through.source in views_by_id:
                        dependencies_by_id[view_id].add(properties.through.source)

        created = ViewList([])
        for strongly_connected in tarjan(dependencies_by_id):
            to_create = [views_by_id[view_id] for view_id in strongly_connected]
            try:
                created_set = self._client.data_modeling.views.apply(to_create)
            except CogniteAPIError:
                self._client.data_modeling.views.delete(created.as_ids())
                raise original_error from None
            created.extend(created_set)
        return created

    @staticmethod
    def _is_auto_retryable(e: CogniteAPIError) -> bool:
        return isinstance(e.extra, dict) and "isAutoRetryable" in e.extra and e.extra["isAutoRetryable"]

    @staticmethod
    def _is_false_not_exists(e: CogniteAPIError, request_views: set[ViewId]) -> bool:
        """This is a bug in the API where it returns a 400 status complaining that a views does not exist,
        even though they are part of the request. This bug is reported to Cognite. The workaround is to do a
        topological sort of the strongly connected views and retry the request.
        """
        if "not exist" not in e.message and 400 <= e.code < 500:
            return False
        results = re.findall(r"'([a-zA-Z0-9_-]+):([a-zA-Z0-9_]+)/([.a-zA-Z0-9_-]+)'", e.message)
        if not results:
            # No view references in the message
            return False
        error_message_views = {ViewId(*result) for result in results}
        return error_message_views.issubset(request_views)

    @overload
    def retrieve(
        self,
        ids: SequenceNotStr[ViewId],
        format: Literal["read"] = "read",
        include_connected: bool = False,
        include_ancestor: bool = False,
    ) -> ViewList: ...

    @overload
    def retrieve(
        self,
        ids: SequenceNotStr[ViewId],
        format: Literal["write"] = "write",
        include_connected: bool = False,
        include_ancestor: bool = False,
    ) -> ViewApplyList: ...

    def retrieve(
        self,
        ids: SequenceNotStr[ViewId],
        format: Literal["read", "write"] = "read",
        include_connected: bool = False,
        include_ancestor: bool = False,
    ) -> ViewList | ViewApplyList:
        if not include_connected and not include_ancestor:
            # Default .as_write() method does not work for views as they include parent properties.
            output = super().retrieve(ids)
        else:
            # Retrieve recursively updates the cache.
            output = self._retrieve_recursive(ids, include_connected, include_ancestor)
        if format == "write":
            return ViewApplyList([self.as_write(view) for view in output])
        return output

    def _retrieve(self, ids: SequenceNotStr[ViewId]) -> ViewList:
        return self._client.data_modeling.views.retrieve(cast(Sequence, ids))

    def _update(self, items: Sequence[ViewApply]) -> ViewList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[ViewId]) -> list[ViewId]:
        return self._client.data_modeling.views.delete(cast(Sequence, ids))

    def _as_write_raw(self, view: View) -> dict[str, Any]:
        dumped = view.as_write().dump()
        if view.properties:
            # All read version of views have all the properties of their parent views.
            # We need to remove these properties to compare with the local view.
            parents = self._retrieve_recursive(view.implements or [], include_connections=False, include_ancestors=True)
            for parent in parents:
                for prop_name, prop in (parent.as_write().properties or {}).items():
                    existing = dumped["properties"].get(prop_name)
                    if existing is None:
                        continue
                    if existing == prop.dump():
                        dumped["properties"].pop(prop_name, None)
                    # If the child overrides the parent, we keep the child's property.

        if "properties" in dumped and not dumped["properties"]:
            # All properties were removed, so we remove the properties key.
            dumped.pop("properties", None)
        return dumped

    def are_equal(self, local: ViewApply, remote: View) -> bool:
        local_dumped = local.dump()
        if not remote.implements:
            return local_dumped == remote.as_write().dump()

        cdf_resource_dumped = self._as_write_raw(remote)

        if "properties" in local_dumped and not local_dumped["properties"]:
            # In case the local properties are set to an empty dict.
            local_dumped.pop("properties", None)

        return local_dumped == cdf_resource_dumped

    def as_write(self, view: View) -> ViewApply:
        return ViewApply.load(self._as_write_raw(view))

    def _retrieve_recursive(
        self, view_ids: SequenceNotStr[ViewId], include_connections: bool, include_ancestors: bool
    ) -> ViewList:
        """Retrieves all views with the

        This will mutate the cache passed in, and return a list of views that are the ancestors
        of the views in the parents list.

        Args:
            view_ids: The views to retrieve.
            include_connections: Whether to include all connected views.
            include_ancestors: Whether to include all ancestors.
        """
        last_batch = set(view_ids)
        found = ViewList([])
        found_ids: set[ViewId] = set()
        while last_batch:
            to_retrieve_from_cdf: set[ViewId] = set()
            batch_ids: set[ViewId] = set()
            for view_id in last_batch:
                if view_id in found_ids:
                    continue
                elif view_id in self._items_by_id:
                    view = self._items_by_id[view_id]
                    found.append(view)
                    found_ids.add(view_id)
                    batch_ids.update(self.get_connected_views(view, include_ancestors, include_connections, found_ids))
                else:
                    to_retrieve_from_cdf.add(view_id)

            if to_retrieve_from_cdf:
                retrieved_batch = self._client.data_modeling.views.retrieve(list(to_retrieve_from_cdf))
                self._items_by_id.update({view.as_id(): view for view in retrieved_batch})
                found.extend(retrieved_batch)
                found_ids.update({view.as_id() for view in retrieved_batch})
                for view in retrieved_batch:
                    batch_ids.update(self.get_connected_views(view, include_ancestors, include_connections, found_ids))
            last_batch = batch_ids

        if self.cache is False:
            # We must update the cache to retrieve recursively.
            # If the cache is disabled, bust the cache to avoid storing the retrieved views.
            self.bust_cache()
        return found

    @staticmethod
    def get_connected_views(
        view: View | ViewApply,
        include_parents: bool = True,
        include_connections: bool = True,
        skip_ids: set[ViewId] | None = None,
    ) -> list[ViewId]:
        connected_ids: set[ViewId] = set(view.implements or []) if include_parents else set()
        if include_connections:
            for prop in (view.properties or {}).values():
                if isinstance(prop, MappedProperty | MappedPropertyApply) and prop.source:
                    connected_ids.add(prop.source)
                elif isinstance(
                    prop, EdgeConnection | EdgeConnectionApply | ReverseDirectRelation | ReverseDirectRelationApply
                ):
                    connected_ids.add(prop.source)

                if isinstance(prop, EdgeConnection | EdgeConnectionApply) and prop.edge_source:
                    connected_ids.add(prop.edge_source)
                elif isinstance(prop, ReverseDirectRelation | ReverseDirectRelationApply) and isinstance(
                    prop.through.source, ViewId
                ):
                    connected_ids.add(prop.through.source)
        if skip_ids:
            return [view_id for view_id in connected_ids if view_id not in skip_ids]
        return list(connected_ids)

    def _create_list(self, items: Sequence[View]) -> ViewList:
        return ViewList(items)

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> ViewApplyList:
        return ViewApplyList(schema.views.values())


class DataModelLoader(DataModelingLoader[DataModelId, DataModelApply, DataModel, DataModelApplyList, DataModelList]):
    resource_name = "data_models"
    dependencies = frozenset({SpaceLoader, ViewLoader})

    @classmethod
    def get_id(cls, item: DataModel | DataModelApply | DataModelId | dict) -> DataModelId:
        if isinstance(item, DataModel | DataModelApply):
            return item.as_id()
        if isinstance(item, dict):
            return DataModelId.load(item)
        return item

    @classmethod
    def merge(cls, local: DataModelApply, remote: DataModel) -> DataModelApply:
        if local.as_id() != remote.as_id():
            raise ValueError(f"Cannot merge data models with different IDs: {local.as_id()} and {remote.as_id()}")
        existing_view = {
            view.as_id() if isinstance(view, View) else view: view.as_write() if isinstance(view, View) else view
            for view in remote.views
        }
        new_views = {view.as_id() if isinstance(view, ViewApply) else view: view for view in local.views or []}
        merged_views: list[ViewId | ViewApply] = []
        for view_id, view in existing_view.items():
            if view_id in new_views:
                merged_views.append(new_views.pop(view_id))
            else:
                merged_views.append(view)
        merged_views.extend(new_views.values())

        return DataModelApply(
            space=local.space,
            external_id=local.external_id,
            version=local.version,
            description=local.description or remote.description,
            name=local.name or remote.name,
            views=merged_views,
        )

    def _create(self, items: Sequence[DataModelApply]) -> DataModelList:
        return self._client.data_modeling.data_models.apply(items)

    def _retrieve(self, ids: SequenceNotStr[DataModelId]) -> DataModelList:
        return self._client.data_modeling.data_models.retrieve(cast(Sequence, ids))

    def _update(self, items: Sequence[DataModelApply]) -> DataModelList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[DataModelId]) -> list[DataModelId]:
        return self._client.data_modeling.data_models.delete(cast(Sequence, ids))

    def _create_list(self, items: Sequence[DataModel]) -> DataModelList:
        return DataModelList(items)

    def are_equal(self, local: DataModelApply, remote: DataModel) -> bool:
        local_dumped = local.dump()
        cdf_resource_dumped = remote.as_write().dump()

        # Data models that have the same views, but in different order, are considered equal.
        # We also account for whether views are given as IDs or View objects.
        local_dumped["views"] = sorted(
            (v if isinstance(v, ViewId) else v.as_id()).as_tuple() for v in local.views or []
        )
        cdf_resource_dumped["views"] = sorted(
            (v if isinstance(v, ViewId) else v.as_id()).as_tuple() for v in remote.views or []
        )

        return local_dumped == cdf_resource_dumped

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> DataModelApplyList:
        return DataModelApplyList([schema.data_model])


class NodeLoader(DataModelingLoader[NodeId, NodeApply, Node, NodeApplyList, NodeList]):
    resource_name = "nodes"
    dependencies = frozenset({SpaceLoader, ContainerLoader, ViewLoader})
    support_merge = False

    @classmethod
    def get_id(cls, item: Node | NodeApply | NodeId | dict) -> NodeId:
        if isinstance(item, Node | NodeApply):
            return item.as_id()
        if isinstance(item, dict):
            return NodeId.load(item)
        return item

    @classmethod
    def merge(cls, local: NodeApply, remote: Node) -> NodeApply:
        raise NotImplementedError("Nodes cannot be merged")

    def _create(self, items: Sequence[NodeApply]) -> NodeList:
        self._client.data_modeling.instances.apply(items)
        return self._retrieve([item.as_id() for item in items])

    def _retrieve(self, ids: SequenceNotStr[NodeId]) -> NodeList:
        return self._client.data_modeling.instances.retrieve(cast(Sequence, ids)).nodes

    def _update(self, items: Sequence[NodeApply]) -> NodeList:
        self._client.data_modeling.instances.apply(items, replace=True)
        return self._retrieve([item.as_id() for item in items])

    def _delete(self, ids: SequenceNotStr[NodeId]) -> list[NodeId]:
        return list(self._client.data_modeling.instances.delete(nodes=cast(Sequence, ids)).nodes)

    def _create_list(self, items: Sequence[Node]) -> NodeList:
        return NodeList(items)

    def are_equal(self, local: NodeApply, remote: Node) -> bool:
        local_dumped = local.dump()

        # Note reading from a container is not supported.
        sources = [
            source_prop_pair.source
            for source_prop_pair in local.sources or []
            if isinstance(source_prop_pair.source, ViewId)
        ]
        if sources:
            try:
                cdf_resource_with_properties = self._client.data_modeling.instances.retrieve(
                    nodes=remote.as_id(), sources=sources
                ).nodes[0]
            except CogniteAPIError:
                # View does not exist, so node does not exist.
                return False
        else:
            cdf_resource_with_properties = remote
        cdf_resource_dumped = cdf_resource_with_properties.as_write().dump()

        if "existingVersion" not in local_dumped:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            local_dumped["existingVersion"] = cdf_resource_dumped.get("existingVersion", None)

        return local_dumped == cdf_resource_dumped

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> NodeApplyList:
        return NodeApplyList(schema.node_types.values())


class DataModelLoaderAPI:
    def __init__(self, client: "NeatClient") -> None:
        self._client = client
        self.spaces = SpaceLoader(client)
        self.views = ViewLoader(client)
        self.containers = ContainerLoader(client)
        self.data_models = DataModelLoader(client)
        self.nodes = NodeLoader(client)
        self._loaders: list[DataModelingLoader] = [
            self.spaces,
            self.views,
            self.containers,
            self.data_models,
            self.nodes,
        ]

    def by_dependency_order(
        self, component: Component | Collection[Component] | None = None
    ) -> list[DataModelingLoader]:
        loader_by_type = {type(loader): loader for loader in self._loaders}
        loader_iterable = (
            loader_by_type[loader_cls]  # type: ignore[index]
            for loader_cls in TopologicalSorter(
                {type(loader): loader.dependencies for loader in self._loaders}  # type: ignore[attr-defined]
            ).static_order()
        )
        if component is None:
            return list(loader_iterable)
        components = {component} if isinstance(component, str) else set(component)
        components = {{"node_type": "nodes"}.get(component, component) for component in components}
        return [loader for loader in loader_iterable if loader.resource_name in components]



"""Validators for checking if defined data model is within CDF DMS schema limits."""

from typing import Literal

from cognite.neat._data_model.models.dms._container import ContainerRequest
from cognite.neat._data_model.models.dms._data_types import EnumProperty, ListablePropertyTypeDefinition
from cognite.neat._data_model.models.dms._indexes import BtreeIndex, InvertedIndex
from cognite.neat._data_model.models.dms._view_property import (
    ViewCorePropertyRequest,
)
from cognite.neat._data_model.rules.dms._base import (
    DataModelRule,
)
from cognite.neat._issues import ConsistencyError, Recommendation

BASE_CODE = "NEAT-DMS-LIMITS"


class DataModelViewCountIsOutOfLimits(DataModelRule):
    """Validates that the data model does not exceed the maximum number of views.

    ## What it does
    This validator checks that the total number of views referenced by the data model
    does not exceed the limit defined in the CDF project.

    ## Why is this bad?
    CDF enforces limits on the number of views per data model to ensure optimal performance
    and resource utilization.

    ## Example
    If the CDF project has a limit of 100 views per data model, and the data model
    references 120 views, this validator will raise a ConsistencyError issue.

    """

    code = f"{BASE_CODE}-DATA-MODEL-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        if self.validation_resources.merged_data_model.views is None:
            errors.append(
                ConsistencyError(
                    message="The data model does not have any views. This means it is not a data model.",
                    code=self.code,
                )
            )

        elif (
            len(self.validation_resources.merged_data_model.views) > self.validation_resources.limits.data_models.views
        ):
            errors.append(
                ConsistencyError(
                    message=(
                        f"The data model references {len(self.validation_resources.merged_data_model.views)} views, "
                        "which exceeds the limit of "
                        f"{self.validation_resources.limits.data_models.views} views per data model."
                    ),
                    code=self.code,
                )
            )
        return errors


### View level limits


class ViewPropertyCountIsOutOfLimits(DataModelRule):
    """Validates that a view does not exceed the maximum number of properties.

    ## What it does
    Checks that the view has no more properties than the CDF limit allows.

    ## Why is this bad?
    CDF enforces limits on the number of properties per view to ensure optimal performance.

    ## Example
    If a view has 150 properties and the CDF limit is 100 properties per view,
    this validator will raise a ConsistencyError issue.
    """

    code = f"{BASE_CODE}-VIEW-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for view_ref in self.validation_resources.merged_data_model.views or []:
            if not (view := self.validation_resources.expand_view_properties(view_ref)):
                raise RuntimeError(f"{type(self).__name__}: View {view_ref!s} not found. This is a bug in NEAT.")

            if view.properties and len(view.properties) > self.validation_resources.limits.views.properties:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"View {view_ref!s} has {len(view.properties)} properties,"
                            " which exceeds the limit of "
                            f"{self.validation_resources.limits.views.properties} properties per view."
                        ),
                        code=self.code,
                    )
                )

            elif not view.properties:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"View {view_ref!s} does "
                            "not have any properties defined, either directly or through implements."
                        ),
                        fix="Define at least one property for view",
                        code=self.code,
                    )
                )

        return errors


class ViewContainerCountIsOutOfLimits(DataModelRule):
    """Validates that a view does not reference too many containers.

    ## What it does
    Checks that the view references no more containers than the CDF limit allows.

    ## Why is this bad?
    Mapping too many containers to a single view can lead to performance issues to increasing number of joins
    that need to be performed when querying data through the view.

    ## Example
    If a view references 20 containers and the CDF limit is 10 containers per view,
    this validator will raise a Recommendation.
    """

    code = f"{BASE_CODE}-VIEW-002"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        # Single loop over all views
        for view_ref in self.validation_resources.merged_data_model.views or []:
            if not (view := self.validation_resources.expand_view_properties(view_ref)):
                raise RuntimeError(f"{type(self).__name__}: View {view_ref!s} not found. This is a bug in NEAT.")

            if view.properties:
                count = len(
                    {
                        prop.container
                        for prop in view.properties.values()
                        if (isinstance(prop, ViewCorePropertyRequest) and prop.container)
                    }
                )
                if count > self.validation_resources.limits.views.containers:
                    recommendations.append(
                        Recommendation(
                            message=(
                                f"View {view_ref!s} references "
                                f"{count} containers, which exceeds a common sense limit of "
                                f"{self.validation_resources.limits.views.containers} containers per view."
                            ),
                            code=self.code,
                        )
                    )

        return recommendations


class ViewImplementsCountIsOutOfLimits(DataModelRule):
    """Validates that a view does not implement too many other views.

    ## What it does
    Checks that the view implements no more views than the CDF limit allows.

    ## Why is this bad?
    CDF enforces limits on the number of implemented views to prevent overly deep inheritance hierarchies.

    ## Example
    If a view implements 15 other views and the CDF limit is 10 implemented views per view,
    this validator will raise a ConsistencyError issue.
    """

    code = f"{BASE_CODE}-VIEW-003"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        # Single loop over all views
        for view_ref in self.validation_resources.merged_data_model.views or []:
            ancestors = self.validation_resources.view_ancestors(view_ref)
            if ancestors and len(ancestors) > self.validation_resources.limits.views.implements:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"View {view_ref!s} implements {len(ancestors)} views,"
                            " which exceeds the limit of"
                            f" {self.validation_resources.limits.views.implements} implemented views per view."
                        ),
                        code=self.code,
                    )
                )
        return errors


### Container level limits


class ContainerPropertyCountIsOutOfLimits(DataModelRule):
    """Validates that a container does not exceed the maximum number of properties.

    ## What it does
    Checks that the container has no more properties than the CDF limit allows.

    ## Why is this bad?
    CDF enforces limits on the number of properties per container to ensure optimal performance
    and prevent PostGres tables that have too many columns.

    ## Example
    If a container has 150 properties and the CDF limit is 100 properties per container,
    this validator will raise a ConsistencyError issue.
    """

    code = f"{BASE_CODE}-CONTAINER-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []
        # Single loop over all containers
        for container_ref in self.validation_resources.merged.containers:
            container = self.validation_resources.select_container(container_ref)

            if not container:
                raise RuntimeError(
                    f"{type(self).__name__}: Container {container_ref!s} "
                    "not found in local resources. This is a bug in NEAT."
                )

            if (
                container.properties
                and len(container.properties) > self.validation_resources.limits.containers.properties()
            ):
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Container {container_ref!s} has {len(container.properties)} properties, "
                            "which exceeds the limit of "
                            f"{self.validation_resources.limits.containers.properties()} properties per container."
                        ),
                        fix="Define at least one property for container",
                        code=self.code,
                    )
                )
            elif not container.properties:
                errors.append(
                    ConsistencyError(
                        message=(f"Container {container_ref!s} does not have any properties defined."),
                        fix="Define at least one property for container",
                        code=self.code,
                    )
                )

        return errors


class ContainerPropertyListSizeIsOutOfLimits(DataModelRule):
    """Validates that container property list sizes do not exceed CDF limits.

    ## What it does
    Checks that list-type properties (max_list_size) do not exceed the appropriate limit based on:
    - Data type (Int32, Int64, DirectRelation, etc.)
    - Presence of btree index
    - Default vs maximum limits

    ## Why is this bad?
    CDF enforces different list size limits for different data types and indexing configurations
    to ensure optimal performance and prevent resource exhaustion.

    ## Example
    If a DirectRelation property has max_list_size=2000 with a btree index, but the limit
    is 1000 for indexed DirectRelations, this validator will raise a ConsistencyError issue.

    ## Note
    Enum properties are skipped as they have a separate 32-value limit checked during read time of data model to neat
    as a SyntaxError check.
    """

    code = f"{BASE_CODE}-CONTAINER-002"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        # Single loop over all containers
        for container_ref in self.validation_resources.merged.containers:
            container = self.validation_resources.select_container(container_ref)

            if not container:
                raise RuntimeError(
                    f"{type(self).__name__}: Container {container_ref!s} "
                    "not found in local resources. This is a bug in NEAT."
                )

            properties_by_index_type = self.container_property_by_index_type(container)

            for property_id, property_ in container.properties.items():
                type_ = property_.type

                # Skip enum properties (have separate 32-value limit)
                if isinstance(type_, EnumProperty):
                    continue

                # Only check listable properties with max_list_size set
                if not isinstance(type_, ListablePropertyTypeDefinition) or type_.max_list_size is None:
                    continue

                has_btree_index = property_id in properties_by_index_type[BtreeIndex.model_fields["index_type"].default]
                limit = self.validation_resources.limits.containers.properties.listable(type_, has_btree_index)

                if type_.max_list_size > limit:
                    errors.append(
                        ConsistencyError(
                            message=(
                                f"Container {container_ref!s} has property {property_id} with list size "
                                f"{type_.max_list_size}, which exceeds the limit of {limit} "
                                f"for data type {type_.__class__.__name__}."
                            ),
                            code=self.code,
                        )
                    )

        return errors

    @staticmethod
    def container_property_by_index_type(container: ContainerRequest) -> dict[Literal["btree", "inverted"], list]:
        """Map container properties to their index types for limit validation.

        Categorizes container properties by their index configuration:
        - "btree": Properties with btree indexes (have stricter list size limits)
        - "inverted": Properties with inverted indexes

        This mapping is used to determine the appropriate list size limit for
        each property based on whether it has a btree index.

        Args:
            container: The container to analyze.

        Returns:
            Dictionary with index type strings as keys and lists of property identifiers
            as values. Returns empty lists for both index types if container has no indexes.
        """

        container_property_by_index_type: dict[Literal["btree", "inverted"], list] = {
            BtreeIndex.model_fields["index_type"].default: [],
            InvertedIndex.model_fields["index_type"].default: [],
        }
        if not container.indexes:
            return container_property_by_index_type

        for index in container.indexes.values():
            if isinstance(index, BtreeIndex):
                container_property_by_index_type[BtreeIndex.model_fields["index_type"].default].extend(index.properties)
            elif isinstance(index, InvertedIndex):
                container_property_by_index_type[InvertedIndex.model_fields["index_type"].default].extend(
                    index.properties
                )

        return container_property_by_index_type



import warnings
import zipfile
from collections.abc import Iterator
from pathlib import Path

import yaml
from pydantic import BaseModel

from cognite.neat._data_model.exporters._base import DMSExporter, DMSFileExporter
from cognite.neat._data_model.models.dms import RequestSchema
from cognite.neat._data_model.models.dms._container import ContainerRequest
from cognite.neat._data_model.models.dms._references import NodeReference
from cognite.neat._data_model.models.dms._views import ViewRequest


class DMSAPIExporter(DMSExporter[RequestSchema]):
    def export(self, data_model: RequestSchema) -> RequestSchema:
        return data_model

    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        raise RuntimeError(f"{type(self).__name__} does not support export_to_file method.")


class DMSAPIYAMLExporter(DMSAPIExporter, DMSFileExporter[RequestSchema]):
    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        """Export the data model to a YAML files or zip file in API format.

        Args:
            data_model: Request schema
            file_path: The directory or zip file to save the schema to.

        """

        if file_path.is_dir():
            self._export_to_directory(data_model, file_path)
        else:
            self._export_to_zip_file(data_model, file_path)

    def _export_to_directory(self, data_model: RequestSchema, directory: Path) -> None:
        """Save the schema to a directory as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            data_model: Request schema
            directory: The directory to save the schema to.
        """

        subdir = directory / "data_models"
        subdir.mkdir(parents=True, exist_ok=True)

        for file_path, yaml_content in self._generate_yaml_entries(data_model):
            full_path = subdir / file_path
            # Create parent directories if needed (e.g., for views/, containers/, nodes/)
            full_path.parent.mkdir(parents=True, exist_ok=True)
            full_path.write_text(
                yaml_content,
                encoding=self.ENCODING,
                newline=self.NEW_LINE,
            )

    def _export_to_zip_file(self, data_model: RequestSchema, zip_file: Path) -> None:
        """Save the schema as a zip file containing a directory as YAML files.
        This is compatible with the Cognite-Toolkit convention.

        Args:
            data_model: Request schema
            zip_file: The zip file path to save the schema to.
        """
        if zip_file.suffix not in {".zip"}:
            warnings.warn("File extension is not .zip, adding it to the file name", stacklevel=2)
            zip_file = zip_file.with_suffix(".zip")

        with zipfile.ZipFile(zip_file, "w") as zip_ref:
            for file_path, yaml_content in self._generate_yaml_entries(data_model):
                zip_ref.writestr(f"data_models/{file_path}", yaml_content)

    def _generate_yaml_entries(self, data_model: RequestSchema) -> Iterator[tuple[str, str]]:
        """Generate file paths and YAML content for all data model components.

        This helper method eliminates duplication by centralizing the logic for
        iterating through spaces, views, containers, and node types.

        Args:
            data_model: Request schema

        Yields:
            Tuples of (file_path, yaml_content) for each component.
            File paths are relative to the data_models directory.
        """

        def _dump(item: BaseModel) -> str:
            return yaml.safe_dump(item.model_dump(mode="json", by_alias=True), sort_keys=False)

        # Export spaces
        for space in data_model.spaces:
            yield f"{space.space}.space.yaml", _dump(space)

        # Export data model
        yield f"{data_model.data_model.external_id}.datamodel.yaml", _dump(data_model.data_model)

        component_configs: list[tuple[str, list[ViewRequest] | list[ContainerRequest] | list[NodeReference]]] = [
            ("views", data_model.views),
            ("containers", data_model.containers),
            ("nodes", data_model.node_types),
        ]

        for dir_prefix, components in component_configs:
            file_suffix = dir_prefix.removesuffix("s")
            for component in components:
                yield f"{dir_prefix}/{component.external_id}.{file_suffix}.yaml", _dump(component)


class DMSAPIJSONExporter(DMSAPIExporter, DMSFileExporter[RequestSchema]):
    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        """Export the data model to a JSON file in API format."""
        if file_path.suffix.lower() != ".json":
            raise ValueError("The file path must have a .json extension.")

        file_path.write_text(
            data_model.model_dump_json(by_alias=True),
            encoding=self.ENCODING,
            newline=self.NEW_LINE,
        )



from collections import Counter

from pydantic import Field, ValidationInfo, field_validator

from cognite.neat._data_model.models.entities import ConceptEntity
from cognite.neat._data_model.models.entities._constants import PREFIX_PATTERN, SUFFIX_PATTERN, VERSION_PATTERN
from cognite.neat._utils.text import humanize_collection

from ._base import ResourceMetadata
from ._concept import Concept


class DataModel(ResourceMetadata):
    space: str = Field(
        description="Id of the space that the concept belongs to.",
        min_length=1,
        max_length=43,
        pattern=PREFIX_PATTERN,
        alias="prefix",
    )
    external_id: str = Field(
        description="External-id of the concept.",
        min_length=1,
        max_length=255,
        pattern=SUFFIX_PATTERN,
        alias="suffix",
    )
    version: str = Field(
        description="Version of the concept.",
        max_length=43,
        pattern=VERSION_PATTERN,
    )

    concepts: list[Concept] = Field(
        description="References to the concepts from where this concept will inherit properties.",
    )

    @field_validator("concepts", mode="after")
    def cannot_have_duplicates(cls, value: list[Concept], info: ValidationInfo) -> list[Concept]:
        concept_ids = [
            ConceptEntity(prefix=concept.space, suffix=concept.external_id, version=concept.version)
            for concept in value
        ]

        counts = Counter(concept_ids)
        duplicates = {concept for concept, count in counts.items() if count > 1}

        if duplicates:
            raise ValueError(f"Duplicate concepts found: {humanize_collection(duplicates)}")

        return value



from collections import Counter
from typing import cast

from pydantic import Field, ValidationInfo, field_validator

from cognite.neat._data_model.models.entities import ConceptEntity
from cognite.neat._data_model.models.entities._constants import PREFIX_PATTERN, SUFFIX_PATTERN, VERSION_PATTERN
from cognite.neat._utils.text import humanize_collection

from ._base import ResourceMetadata
from ._property import Property


class Concept(ResourceMetadata):
    space: str = Field(
        description="Id of the space that the concept belongs to.",
        min_length=1,
        max_length=43,
        pattern=PREFIX_PATTERN,
        alias="prefix",
    )
    external_id: str = Field(
        description="External-id of the concept.",
        min_length=1,
        max_length=255,
        pattern=SUFFIX_PATTERN,
        alias="suffix",
    )
    version: str | None = Field(
        default=None,
        description="Version of the concept.",
        max_length=43,
        pattern=VERSION_PATTERN,
    )

    implements: list[ConceptEntity] | None = Field(
        default=None,
        description="References to the concepts from where this concept will inherit properties.",
    )

    properties: dict[str, Property] | None = Field(default=None, description="Properties associated with the concept.")

    @field_validator("implements", mode="after")
    def cannot_implement_itself(cls, value: list[ConceptEntity], info: ValidationInfo) -> list[ConceptEntity]:
        if not value:
            return value

        this_concept = ConceptEntity(
            prefix=info.data["space"],
            suffix=info.data["external_id"],
            version=cast(str, info.data.get("version")),
        )

        if this_concept in value:
            raise ValueError("A concept cannot implement itself.")

        return value

    @field_validator("implements", mode="after")
    def cannot_have_duplicates(cls, value: list[ConceptEntity], info: ValidationInfo) -> list[ConceptEntity]:
        counts = Counter(value)
        duplicates = {concept for concept, count in counts.items() if count > 1}

        if duplicates:
            raise ValueError(f"Duplicate concepts found: {humanize_collection(duplicates)}")

        return value



from typing import Generic, TypeVar

from pydantic import BaseModel, Field

from cognite.neat._utils.useful_types import BaseModelObject

T = TypeVar("T", bound=BaseModel)


class PagedResponse(BaseModel, Generic[T]):
    items: list[T]
    next_cursor: str | None = Field(None, alias="nextCursor")


class ResourceLimit(BaseModel):
    """Model for resources with count and limit."""

    count: int
    limit: int


class InstancesDetail(BaseModel, populate_by_name=True):
    """Model for instances with detailed metrics."""

    edges: int
    soft_deleted_edges: int = Field(alias="softDeletedEdges")
    nodes: int
    soft_deleted_nodes: int = Field(alias="softDeletedNodes")
    instances: int
    instances_limit: int = Field(alias="instancesLimit")
    soft_deleted_instances: int = Field(alias="softDeletedInstances")
    soft_deleted_instances_limit: int = Field(alias="softDeletedInstancesLimit")


class StatisticsResponse(BaseModel, populate_by_name=True):
    """Main API response model."""

    spaces: ResourceLimit
    containers: ResourceLimit
    views: ResourceLimit
    data_models: ResourceLimit = Field(alias="dataModels")
    container_properties: ResourceLimit = Field(alias="containerProperties")
    instances: InstancesDetail
    concurrent_read_limit: int = Field(alias="concurrentReadLimit")
    concurrent_write_limit: int = Field(alias="concurrentWriteLimit")
    concurrent_delete_limit: int = Field(alias="concurrentDeleteLimit")


class SpaceStatisticsItem(BaseModelObject, populate_by_name=True):
    """Individual space statistics item."""

    space: str
    containers: int
    views: int
    data_models: int
    edges: int
    soft_deleted_edges: int
    nodes: int
    soft_deleted_nodes: int

    @property
    def is_empty(self) -> bool:
        """Check if the space has zero usage."""
        return (
            self.containers == 0 and self.views == 0 and self.data_models == 0 and self.edges == 0 and self.nodes == 0
        )


class SpaceStatisticsResponse(BaseModelObject, populate_by_name=True):
    """Response model for space statistics endpoint."""

    items: list[SpaceStatisticsItem]

    def empty_spaces(self) -> list[str]:
        """Get a list of space identifiers that have zero usage."""
        return [item.space for item in self.items if item.is_empty]



from cognite.neat._data_model.models.dms import (
    BtreeIndex,
    ConstraintDefinition,
    ContainerPropertyDefinition,
    ContainerRequest,
    EnumProperty,
    EnumValue,
    FloatProperty,
    IndexDefinition,
    ListablePropertyTypeDefinition,
    PropertyTypeDefinition,
    RequiresConstraintDefinition,
    TextProperty,
    UniquenessConstraintDefinition,
)
from cognite.neat._data_model.models.dms._data_types import Unit

from ._differ import ItemDiffer, ObjectDiffer, field_differences
from .data_classes import (
    AddedField,
    ChangedField,
    FieldChange,
    FieldChanges,
    RemovedField,
    SeverityType,
)


class ContainerDiffer(ItemDiffer[ContainerRequest]):
    def diff(self, current: ContainerRequest, new: ContainerRequest) -> list[FieldChange]:
        changes = self._diff_name_description(current, new)

        if current.used_for != new.used_for:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path("usedFor"),
                    current_value=current.used_for,
                    new_value=new.used_for,
                )
            )
        changes.extend(
            field_differences(
                "properties",
                current.properties,
                new.properties,
                add_severity=SeverityType.SAFE,
                remove_severity=SeverityType.BREAKING,
                differ=ContainerPropertyDiffer("properties"),
            )
        )
        changes.extend(
            # MyPy fails to understand that ConstraintDefinition and Constraint are compatible here
            field_differences(  # type: ignore[misc]
                "constraints",
                current.constraints,
                new.constraints,
                add_severity=SeverityType.WARNING,
                remove_severity=SeverityType.WARNING,
                differ=ConstraintDiffer("constraints"),
            )
        )
        changes.extend(
            # MyPy fails to understand that IndexDefinition and Index are compatible here
            field_differences(  # type: ignore[misc]
                "indexes",
                current.indexes,
                new.indexes,
                add_severity=SeverityType.SAFE,
                remove_severity=SeverityType.WARNING,
                differ=IndexDiffer("indexes"),
            )
        )

        return changes


class ContainerPropertyDiffer(ObjectDiffer[ContainerPropertyDefinition]):
    def diff(
        self, current: ContainerPropertyDefinition, new: ContainerPropertyDefinition, identifier: str
    ) -> list[FieldChange]:
        changes = self._diff_name_description(current, new, identifier=identifier)
        diffs = DataTypeDiffer(self._get_path(f"{identifier}.type")).diff(current.type, new.type)
        if diffs:
            changes.append(FieldChanges(field_path=self._get_path(f"{identifier}.type"), changes=diffs))

        if current.immutable != new.immutable:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.WARNING,
                    field_path=self._get_path(f"{identifier}.immutable"),
                    current_value=current.immutable,
                    new_value=new.immutable,
                )
            )

        if current.nullable != new.nullable:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path(f"{identifier}.nullable"),
                    current_value=current.nullable,
                    new_value=new.nullable,
                )
            )
        if current.auto_increment != new.auto_increment:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.WARNING,
                    field_path=self._get_path(f"{identifier}.autoIncrement"),
                    current_value=current.auto_increment,
                    new_value=new.auto_increment,
                )
            )

        if current.default_value != new.default_value:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.WARNING,
                    field_path=self._get_path(f"{identifier}.defaultValue"),
                    current_value=str(current.default_value),
                    new_value=str(new.default_value),
                )
            )
        return changes


class ConstraintDiffer(ObjectDiffer[ConstraintDefinition]):
    def diff(self, current: ConstraintDefinition, new: ConstraintDefinition, identifier: str) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.constraint_type != new.constraint_type:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path(f"{identifier}.constraintType"),
                    current_value=current.constraint_type,
                    new_value=new.constraint_type,
                )
            )
        if (
            isinstance(current, RequiresConstraintDefinition)
            and isinstance(new, RequiresConstraintDefinition)
            and current.require != new.require
        ):
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path(f"{identifier}.require"),
                    current_value=str(current.require),
                    new_value=str(new.require),
                )
            )
        elif isinstance(current, UniquenessConstraintDefinition) and isinstance(new, UniquenessConstraintDefinition):
            # The order of the properties matter.
            if current.properties != new.properties:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.BREAKING,
                        field_path=self._get_path(f"{identifier}.properties"),
                        current_value=str(current.properties),
                        new_value=str(new.properties),
                    )
                )
            if current.by_space != new.by_space:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.BREAKING,
                        field_path=self._get_path(f"{identifier}.bySpace"),
                        current_value=current.by_space,
                        new_value=new.by_space,
                    )
                )
            return changes
        return changes


class IndexDiffer(ObjectDiffer[IndexDefinition]):
    def diff(self, current: IndexDefinition, new: IndexDefinition, identifier: str) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.index_type != new.index_type:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path(f"{identifier}.indexType"),
                    current_value=current.index_type,
                    new_value=new.index_type,
                )
            )
        else:
            if current.properties != new.properties:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.BREAKING,
                        field_path=self._get_path(f"{identifier}.properties"),
                        current_value=str(current.properties),
                        new_value=str(new.properties),
                    )
                )
        if isinstance(current, BtreeIndex) and isinstance(new, BtreeIndex):
            if current.cursorable != new.cursorable:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.BREAKING,
                        field_path=self._get_path(f"{identifier}.cursorable"),
                        current_value=current.cursorable,
                        new_value=new.cursorable,
                    )
                )
            if current.by_space != new.by_space:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.BREAKING,
                        field_path=self._get_path(f"{identifier}.bySpace"),
                        current_value=current.by_space,
                        new_value=new.by_space,
                    )
                )
        return changes


class DataTypeDiffer(ItemDiffer[PropertyTypeDefinition]):
    def diff(self, current: PropertyTypeDefinition, new: PropertyTypeDefinition) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.type != new.type:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path("type"),
                    current_value=current.type,
                    new_value=new.type,
                )
            )
        if isinstance(current, ListablePropertyTypeDefinition) and isinstance(new, ListablePropertyTypeDefinition):
            changes.extend(self._diff_listable_property(current, new))

        if isinstance(current, TextProperty) and isinstance(new, TextProperty):
            changes.extend(self._diff_text_property(current, new))

        if isinstance(current, FloatProperty) and isinstance(new, FloatProperty):
            changes.extend(self._diff_float_unit(current.unit, new.unit))

        if isinstance(current, EnumProperty) and isinstance(new, EnumProperty):
            changes.extend(self._diff_enum_property(current, new))

        return changes

    def _diff_listable_property(
        self, current: ListablePropertyTypeDefinition, new: ListablePropertyTypeDefinition
    ) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.list != new.list:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path("list"),
                    current_value=current.list,
                    new_value=new.list,
                )
            )
        if current.max_list_size != new.max_list_size:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING
                    if new.max_list_size is not None
                    and current.max_list_size is not None
                    and new.max_list_size < current.max_list_size
                    else SeverityType.WARNING,
                    field_path=self._get_path("maxListSize"),
                    current_value=current.max_list_size,
                    new_value=new.max_list_size,
                )
            )
        return changes

    def _diff_float_unit(self, current: Unit | None, new: Unit | None) -> list[FieldChange]:
        if current is not None and new is None:
            return [
                RemovedField(
                    field_path=self._get_path("unit"), item_severity=SeverityType.WARNING, current_value=current
                )
            ]
        elif current is None and new is not None:
            return [AddedField(field_path=self._get_path("unit"), item_severity=SeverityType.WARNING, new_value=new)]
        elif current is not None and new is not None:
            changes: list[FieldChange] = []
            if current.external_id != new.external_id:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.WARNING,
                        field_path=self._get_path("unit.externalId"),
                        current_value=current.external_id,
                        new_value=new.external_id,
                    )
                )
            if current.source_unit != new.source_unit:
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.WARNING,
                        field_path=self._get_path("unit.sourceUnit"),
                        current_value=current.source_unit,
                        new_value=new.source_unit,
                    )
                )
            if changes:
                return [FieldChanges(field_path=self._get_path("unit"), changes=changes)]
        return []  # No changes

    def _diff_text_property(self, current: TextProperty, new: TextProperty) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.max_text_size != new.max_text_size:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING
                    if new.max_text_size is not None
                    and current.max_text_size is not None
                    and new.max_text_size < current.max_text_size
                    else SeverityType.WARNING,
                    field_path=self._get_path("maxTextSize"),
                    current_value=current.max_text_size,
                    new_value=new.max_text_size,
                )
            )
        if current.collation != new.collation:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.BREAKING,
                    field_path=self._get_path("collation"),
                    current_value=current.collation,
                    new_value=new.collation,
                )
            )
        return changes

    def _diff_enum_property(self, current: EnumProperty, new: EnumProperty) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.unknown_value != new.unknown_value:
            changes.append(
                ChangedField(
                    item_severity=SeverityType.WARNING,
                    field_path=self._get_path("unknownValue"),
                    current_value=current.unknown_value,
                    new_value=new.unknown_value,
                )
            )
        changes.extend(
            field_differences(
                self._get_path("values"),
                current.values,
                new.values,
                add_severity=SeverityType.SAFE,
                remove_severity=SeverityType.BREAKING,
                differ=EnumValueDiffer(self._get_path("values")),
            )
        )
        return changes


class EnumValueDiffer(ObjectDiffer[EnumValue]):
    def diff(self, current: EnumValue, new: EnumValue, identifier: str) -> list[FieldChange]:
        return self._diff_name_description(current, new, identifier=identifier)



"""Validators for checking performance-related aspects of the data model."""

from cognite.neat._data_model._analysis import RequiresChangeStatus
from cognite.neat._data_model._constants import COGNITE_SPACES
from cognite.neat._data_model.models.dms._data_types import DirectNodeRelation
from cognite.neat._data_model.models.dms._indexes import BtreeIndex
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._issues import Recommendation

BASE_CODE = "NEAT-DMS-PERFORMANCE"


class MissingRequiresConstraint(DataModelRule):
    """
    Recommends adding requires constraints to optimize query performance.

    ## What it does
    Identifies views that is mapping to containers where adding a requires constraint,
    would improve query performance. The recommendation message indicates whether the
    change is "safe" or requires attention to potential ingestion dependencies.

    ## Why is this important?
    Views without proper requires constraints may have poor query performance.
    Adding requires constraints enables queries to perform under-the-hood optimizations.

    ## Example
    View `Valve` is mapping to both containers `Valve` and `CogniteEquipment`.
    A `requires` constraint from `Valve` to `CogniteEquipment` is likely needed
    to enable efficient query performance.
    """

    code = f"{BASE_CODE}-001"
    issue_type = Recommendation
    alpha = True

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged.views:
            changes = self.validation_resources.get_requires_changes_for_view(view_ref)

            if changes.status != RequiresChangeStatus.CHANGES_AVAILABLE:
                continue

            for src, dst in changes.to_add:
                src_views = self.validation_resources.views_by_container.get(src, set())
                other_views_with_src = src_views - {view_ref}
                views_impacted_by_change = self.validation_resources.find_views_mapping_to_containers([src, dst])

                # Check if this is a "safe" recommendation (no cross-view dependencies)
                is_safe = not other_views_with_src or view_ref in views_impacted_by_change

                message = (
                    f"View '{view_ref!s}' is not optimized for querying. "
                    f"Add a 'requires' constraint from the container '{src!s}' to '{dst!s}'."
                )
                if not is_safe:
                    # Find a superset view to suggest for ingestion
                    merged_views = set(self.validation_resources.merged.views)
                    superset_views = views_impacted_by_change & merged_views
                    view_example = min(superset_views, key=str) if superset_views else None
                    message += (
                        " Note: this causes an ingestion dependency for this view, "
                        " if you will be using this view to ingest instances, you will "
                        f"need to populate these instances into '{dst!s}' first"
                        + (f", for example through view '{view_example!s}'." if view_example else ".")
                    )

                recommendations.append(
                    Recommendation(
                        message=message,
                        fix="Add requires constraint between the containers",
                        code=self.code,
                    )
                )

        return recommendations


class SuboptimalRequiresConstraint(DataModelRule):
    """
    Recommends removing requires constraints that are not optimal.

    ## What it does
    Identifies existing requires constraints that are not optimal for querying purposes,
    as they are either redundant or create unnecessary ingestion dependencies when all
    other optimal constraints are applied. These constraints can be safely removed
    without affecting query performance.

    ## Why is this important?
    Unnecessary requires constraints can:
    - Create unnecessary ingestion dependencies
    - Cause invalid requires constraint cycles if optimal constraints are applied

    ## Example
    Container `Tag` has a `requires` constraint to `Pump`, but NEAT determined that
    `Pump → Tag` is more optimal. The existing `Tag → Pump` constraint should then
    be removed when applying all optimal constraints.
    """

    code = f"{BASE_CODE}-002"
    issue_type = Recommendation
    alpha = True

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged.views:
            changes = self.validation_resources.get_requires_changes_for_view(view_ref)

            if changes.status != RequiresChangeStatus.CHANGES_AVAILABLE:
                continue

            for src, dst in changes.to_remove:
                recommendations.append(
                    Recommendation(
                        message=(
                            f"View '{view_ref!s}' is mapping to container '{src!s}' "
                            f"that has a requires constraint to '{dst!s}'. This constraint is "
                            "not part of the optimal structure. Consider removing it."
                        ),
                        fix="Remove the unnecessary requires constraint",
                        code=self.code,
                    )
                )

        return recommendations


class UnresolvableQueryPerformance(DataModelRule):
    """
    Identifies views with query performance issues that cannot be resolved.
    This is likely to be caused by unintended modeling choices.

    ## What it does
    Detects views where no valid requires constraint solution exists:

    1. **View maps only to CDF built-in containers**:
       Since CDF containers cannot be modified, no requires can be added.

    2. **No valid solution**:
       This view is causing issues when optimizing requires constraints
       for other views, due to its structure (mapping non-overlapping containers)

    ## Why is this important?
    These views will have suboptimal query performance that CANNOT be fixed by
    adding or removing requires constraints. The only solutions require restructuring:
    - Add a view-specific container that requires all the other containers in the view
    - Restructure the view to use different containers

    ## Example
    View `MultipleEquipments` maps only to containers `Valve` and `InstrumentEquipment`.
    The optimal constraints are `Valve → CogniteEquipment` and `InstrumentEquipment → CogniteEquipment`
    due to other views needing these constraints to optimize their query performance.
    This means however, that neither Valve nor InstrumentEquipment can reach each other without
    creating complex ingestion dependencies. The view needs a new container or restructuring.
    """

    code = f"{BASE_CODE}-003"
    issue_type = Recommendation
    alpha = True

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged.views:
            if view_ref.space in COGNITE_SPACES:
                continue

            changes = self.validation_resources.get_requires_changes_for_view(view_ref)

            if changes.status == RequiresChangeStatus.NO_MODIFIABLE_CONTAINERS:
                recommendations.append(
                    Recommendation(
                        message=(
                            f"View '{view_ref!s}' is not optimized for querying. It maps only to CDF "
                            "built-in containers and some of these lack requires constraints between them. "
                            "Consider adding a view-specific container that requires the others."
                        ),
                        fix="Add a container that requires the others, or restructure the view",
                        code=self.code,
                    )
                )
            elif changes.status == RequiresChangeStatus.UNSOLVABLE:
                recommendations.append(
                    Recommendation(
                        message=(
                            f"View '{view_ref!s}' is not optimized for querying. "
                            "No valid requires constraint solution was found. "
                            "Consider adding a view-specific container that requires the others."
                        ),
                        fix="Add a container that requires the others, or restructure the view",
                        code=self.code,
                    )
                )

        return recommendations


class MissingReverseDirectRelationTargetIndex(DataModelRule):
    """
    Recommends adding a cursorable index on direct relation properties that are
    targets of reverse direct relations for query performance.

    ## What it does
    Identifies direct relation properties that are referenced by reverse direct relations
    but lack a cursorable B-tree index. When querying through a reverse direct relation,
    CDF needs to look up nodes that have the direct relation pointing to the
    source nodes. Without an index, this requires scanning many nodes inefficiently.

    ## Why is this important?
    Traversing a reverse direct relation (inwards direction) requires an index on the
    target direct relation property. Without this index, queries will be inefficient,
    potentially leading to timeouts over time, as they won't scale well with data volume.

    The exception is when the target direct relation is a list property. In that case,
    this validator will not flag them, as reverse direct relations targeting lists of
    direct relations needs to be traversed using the `instances/search` endpoint instead,
    which does not directly benefit from adding indexes to container properties.

    ## Example
    View `WindFarm` has a reverse property `turbines` through `WindTurbine.windFarm`.
    Container `WindTurbine` should have a cursorable B-tree index on the `windFarm`
    property to enable efficient traversal from WindFarm to its turbines.
    """

    code = f"{BASE_CODE}-004"
    issue_type = Recommendation
    alpha = True

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            # Skip if container or container property couldn't be resolved
            if not resolved.container or not resolved.container_property:
                continue

            # Skip CDM containers - we can't modify these
            if resolved.container_ref.space in COGNITE_SPACES:
                continue

            # Must be a DirectNodeRelation type (other types handled by ReverseConnectionContainerPropertyWrongType)
            if not isinstance(resolved.container_property.type, DirectNodeRelation):
                continue

            # Skip if this is a list direct relation - indexes are not supported for list properties
            if resolved.container_property.type.list:
                continue

            # Skip if there's already a cursorable B-tree index on this property
            if resolved.container.indexes and any(
                isinstance(index, BtreeIndex)
                and index.cursorable
                and resolved.container_property_id in index.properties
                for index in resolved.container.indexes.values()
            ):
                continue

            recommendations.append(
                Recommendation(
                    message=(
                        f"View '{resolved.reverse_view_ref!s}' has a reverse direct relation "
                        f"'{resolved.reverse_property_id}' that cannot be efficently traversed through queries, "
                        f"since it points to container {resolved.container_ref!s}' "
                        f"property '{resolved.container_property_id}' that is unindexed. "
                    ),
                    fix="Add a cursorable B-tree index on the target direct relation property",
                    code=self.code,
                )
            )

        return recommendations



from collections.abc import Mapping
from typing import Annotated, Any, Literal, cast, get_args

from pydantic import (
    AliasGenerator,
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    field_validator,
    model_validator,
)
from pydantic.alias_generators import to_camel
from pydantic.fields import FieldInfo

from cognite.neat._data_model.models.entities import ParsedEntity, parse_entities, parse_entity
from cognite.neat._utils.text import title_case
from cognite.neat._utils.useful_types import CellValueType

# This marker is used to identify creator in the description field.
CREATOR_MARKER = "Creator: "
CREATOR_KEY = "creator"


def parse_entity_str(v: str) -> ParsedEntity:
    if isinstance(v, ParsedEntity):
        return v
    try:
        return parse_entity(v)
    except ValueError as e:
        raise ValueError(f"Invalid entity syntax: {e}") from e


def parse_entities_str(v: str) -> list[ParsedEntity] | None:
    if isinstance(v, list) and all(isinstance(item, ParsedEntity) for item in v):
        return v
    try:
        return parse_entities(v)
    except ValueError as e:
        raise ValueError(f"Invalid entity list syntax: {e}") from e


Entity = Annotated[ParsedEntity, BeforeValidator(parse_entity_str, str), PlainSerializer(func=str)]
EntityList = Annotated[
    list[ParsedEntity],
    BeforeValidator(parse_entities_str, str),
    PlainSerializer(func=lambda v: ",".join([str(item) for item in v])),
]


class TableObj(
    BaseModel,
    extra="ignore",
    alias_generator=AliasGenerator(
        alias=to_camel,
        validation_alias=title_case,
        serialization_alias=title_case,
    ),
    populate_by_name=True,
): ...


class MetadataValue(TableObj):
    key: str
    value: CellValueType

    @field_validator("key", mode="after")
    def _legacy_external_id(cls, value: str) -> str:
        return "externalId" if value.lower() == "external_id" else value


class DMSProperty(TableObj):
    view: Entity
    view_property: str
    name: str | None = None
    description: str | None = None
    connection: Entity | None
    value_type: Entity
    min_count: int | None
    max_count: int | None
    immutable: bool | None = None
    default: CellValueType | None = None
    auto_increment: bool | None = None
    container: Entity | None = None
    container_property: str | None = None
    container_property_name: str | None = None
    container_property_description: str | None = None
    index: EntityList | None = None
    constraint: EntityList | None = None

    @field_validator("max_count", mode="before")
    @classmethod
    def _legacy_max_count(cls, value: Any) -> Any | None:
        """Validates and converts the max_count field if it uses the legacy 'inf' value."""
        if isinstance(value, str) and value.lower() == "inf":
            return None
        return value

    @model_validator(mode="before")
    def _legacy_cardinality(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Converts Is List to Max Count and Nullable to Min Count."""
        if not data:
            return data

        if "Max Count" not in data and "Is List" in data:
            is_list = data.pop("Is List")
            if isinstance(is_list, bool):
                data["Max Count"] = None if is_list else 1

        if "Min Count" not in data and "Nullable" in data:
            nullable = data.pop("Nullable")
            if isinstance(nullable, bool):
                data["Min Count"] = 0 if nullable else 1

        return data

    @model_validator(mode="after")
    def _legacy_index(self) -> "DMSProperty":
        """Converts Is List to Max Count and Nullable to Min Count."""
        if not self.index:
            return self

        for index in self.index:
            if not index.prefix:
                index.prefix = "inverted" if not self.max_count or self.max_count > 1 else "btree"

        return self

    @model_validator(mode="after")
    def _legacy_constraint(self) -> "DMSProperty":
        """Converts Is List to Max Count and Nullable to Min Count."""
        if not self.constraint:
            return self

        for constraint in self.constraint:
            if not constraint.prefix:
                constraint.prefix = "uniqueness"

        return self


class EntityTableFilter(BaseModel):
    """These are special formats that Neat Table format supports for filters."""

    type: Literal["hasData", "nodeType"]
    entities: EntityList

    def __str__(self) -> str:
        entities_str = ",".join(str(entity) for entity in self.entities)
        return f"{self.type}({entities_str})"


class RAWFilterTableFilter(BaseModel):
    """This is a generic filter that holds raw JSON filter."""

    type: Literal["rawFilter"] = "rawFilter"
    filter: str

    def __str__(self) -> str:
        return f"rawFilter({self.filter})"


def _parse_table_filter(v: str) -> dict[str, str] | EntityTableFilter | RAWFilterTableFilter:
    if isinstance(v, EntityTableFilter | RAWFilterTableFilter):
        return v
    filter_configs = {
        "hasdata(": ("hasData", "entities"),
        "nodetype(": ("nodeType", "entities"),
        "rawfilter(": ("rawFilter", "filter"),
    }
    v_lowered = v.casefold()
    for prefix, (filter_type, field_name) in filter_configs.items():
        if v_lowered.startswith(prefix) and v_lowered.endswith(")"):
            return {"type": filter_type, field_name: v[len(prefix) : -1]}
    # Fallback to raw filter with the whole string
    return {"type": "rawFilter", "filter": v}


TableViewFilter = Annotated[
    EntityTableFilter | RAWFilterTableFilter,
    Field(discriminator="type"),
    BeforeValidator(_parse_table_filter, str),
    PlainSerializer(func=str),
]


class DMSView(TableObj):
    view: Entity
    name: str | None = None
    description: str | None = None
    implements: EntityList | None = None
    filter: TableViewFilter | None = None
    in_model: bool | None = Field(None, exclude=True, description="Legacy column")


class DMSContainer(TableObj):
    container: Entity
    name: str | None = None
    description: str | None = None
    constraint: EntityList | None = None
    used_for: str | None = None

    @model_validator(mode="after")
    def _legacy_constraint(self) -> "DMSContainer":
        """Converts legacy constraint formats to the current standard."""
        if not self.constraint:
            return self

        for constraint in self.constraint:
            # Skip if already in correct format or being wrong but not legacy
            if constraint.prefix == "requires" or constraint.properties:
                continue

            # This part handles legacy constraints
            if not constraint.prefix:
                constraint.prefix = "requires"
                constraint.properties = {"require": constraint.suffix}

            else:
                container_space = constraint.prefix
                container_external_id = constraint.suffix
                constraint.prefix = "requires"
                constraint.suffix = f"{container_space}_{container_external_id}"
                constraint.properties = {"require": f"{container_space}:{container_external_id}"}

        return self


class DMSEnum(TableObj):
    collection: str
    value: str
    name: str | None = None
    description: str | None = None


class DMSNode(TableObj):
    node: Entity


class TableDMS(TableObj):
    metadata: list[MetadataValue]
    properties: list[DMSProperty]
    views: list[DMSView]
    containers: list[DMSContainer] = Field(default_factory=list)
    enum: list[DMSEnum] = Field(default_factory=list)
    nodes: list[DMSNode] = Field(default_factory=list)

    @model_validator(mode="before")
    def _title_case_keys(
        cls, data: dict[str, list[dict[str, CellValueType]]]
    ) -> dict[str, list[dict[str, CellValueType]]]:
        if isinstance(data, dict):
            # We are case-insensitive on the table names.
            return {title_case(k): v for k, v in data.items()}
        return data

    @model_validator(mode="after")
    def _drop_in_model_false_views_definitions(self) -> "TableDMS":
        """These method is used to drop definition of legacy views which have In Model column set to False
        We need to drop these views from Views sheet and also drop their properties from Properties sheet.
        """

        views_to_drop: set[Entity] = set()
        for view in self.views:
            if isinstance(view.in_model, bool) and not view.in_model:
                views_to_drop.add(view.view)

        if not views_to_drop:
            return self

        print(
            "You are using legacy `In Model` column which is no longer in use!"
            f"\nTotal of {len(views_to_drop)} views has `In Model` set to False."
            f"\nThese views and their property definitions will be dropped from the session!"
        )

        self.views = [view for view in self.views if view.view not in views_to_drop]
        self.properties = [prop for prop in self.properties if prop.view not in views_to_drop]

        return self

    @classmethod
    def get_sheet_columns(
        cls, sheet_id: str, sheet: FieldInfo | None = None, *, column_type: Literal["all", "required"] = "required"
    ) -> list[str]:
        if sheet_id not in cls.model_fields.keys():
            raise KeyError(f"Invalid field id: {sheet_id}")
        if sheet is None:
            sheet = cls.model_fields[sheet_id]
        return [
            # We know all fields has validation_alias because of the alias_generator in TableDMS
            cast(str, sheet_field.validation_alias)
            # All the fields in the sheet's model are lists.
            for sheet_field in get_args(sheet.annotation)[0].model_fields.values()
            if sheet_field.is_required() or column_type == "all"
        ]

    @classmethod
    def get_sheet_column_by_name(
        cls, sheet_name: str, *, column_type: Literal["all", "required"] = "required"
    ) -> list[str]:
        for field_id, field_ in cls.model_fields.items():
            if cast(str, field_.validation_alias) == sheet_name:
                return cls.get_sheet_columns(field_id, field_, column_type=column_type)
        raise KeyError(f"Invalid field alias: {sheet_name}")

    @classmethod
    def required_sheets(cls) -> set[str]:
        return {cast(str, field_.validation_alias) for field_ in cls.model_fields.values() if field_.is_required()}


DMS_API_MAPPING: Mapping[str, Mapping[str, str]] = {
    "Views": {
        "space": "View",
        "externalId": "View",
        "version": "View",
        **{
            cast(str, field_.alias): cast(str, field_.validation_alias)
            for field_id, field_ in DMSView.model_fields.items()
            if field_id != "View"
        },
    },
    "Containers": {
        "space": "Container",
        "externalId": "Container",
        **{
            cast(str, field_.alias): cast(str, field_.validation_alias)
            for field_id, field_ in DMSContainer.model_fields.items()
            if field_id != "Container"
        },
    },
    "Properties": {
        "space": "View",
        "externalId": "View",
        "property": "ViewProperty",
        "type": "Value Type",
        **{
            cast(str, field_.alias): cast(str, field_.validation_alias)
            for field_id, field_ in DMSProperty.model_fields.items()
            if field_id not in ("View", "ViewProperty")
        },
    },
}



from typing import Any, Literal

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class BaseModelRequest(BaseModel):
    """Base class for all object. This includes resources and nested objects."""


class Filter(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, extra="ignore", populate_by_name=True)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the object to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.

        """
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True)


class DataModelingFilter(Filter):
    space: str | None = None
    include_global: bool | None = None


class ContainerFilter(DataModelingFilter):
    used_for: Literal["node", "edge", "record", "all"] | None = None


class ViewFilter(DataModelingFilter):
    include_inherited_properties: bool | None = None
    all_versions: bool | None = None


class DataModelFilter(DataModelingFilter):
    inline_views: bool | None = None
    all_versions: bool | None = None



from cognite.neat._data_model.models.dms import (
    DataModelRequest,
)

from ._differ import ItemDiffer
from .data_classes import (
    AddedField,
    ChangedField,
    FieldChange,
    RemovedField,
    SeverityType,
)


class DataModelDiffer(ItemDiffer[DataModelRequest]):
    def diff(self, current: DataModelRequest, new: DataModelRequest) -> list[FieldChange]:
        changes: list[FieldChange] = self._diff_name_description(current, new)
        if current.views != new.views:
            # Added views
            current_views = set(current.views or [])
            for new_view in new.views or []:
                if new_view not in current_views:
                    changes.append(
                        AddedField(
                            item_severity=SeverityType.SAFE,
                            field_path="views",
                            new_value=new_view,
                        )
                    )

            # Removed views
            new_views = set(new.views or [])
            for current_view in current.views or []:
                if current_view not in new_views:
                    changes.append(
                        RemovedField(
                            item_severity=SeverityType.BREAKING,
                            field_path="views",
                            current_value=current_view,
                        )
                    )

            if not changes:
                # If there are no added or removed views, it means the order has changed
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.SAFE,
                        field_path="views",
                        current_value=str(current.views),
                        new_value=str(new.views),
                    )
                )

        return changes



from typing import Any, Literal, TypeAlias

from . import static, templates

ENCODING = "utf-8"

Template: TypeAlias = Literal["issues", "deployment", "statistics"]


def render(template_name: Literal["issues", "deployment", "statistics"], variables: dict[str, Any]) -> str:
    """Generate HTML content from a template and variables."""

    if template_name not in ["issues", "deployment", "statistics"]:
        raise ValueError(f"Unknown template name: {template_name}")

    variables["SHARED_CSS"] = static.shared_style.read_text(encoding=ENCODING)

    if template_name == "issues":
        template = templates.issues.read_text(encoding=ENCODING)
        variables["SCRIPTS"] = static.issues_scripts.read_text(encoding=ENCODING)
        variables["SPECIFIC_CSS"] = static.issues_style.read_text(encoding=ENCODING)

    elif template_name == "deployment":
        template = templates.deployment.read_text(encoding=ENCODING)
        variables["SCRIPTS"] = static.deployment_scripts.read_text(encoding=ENCODING)
        variables["SPECIFIC_CSS"] = static.deployment_style.read_text(encoding=ENCODING)

    elif template_name == "statistics":
        template = templates.statistics.read_text(encoding=ENCODING)
        variables["SCRIPTS"] = static.statistics_scripts.read_text(encoding=ENCODING)
        variables["SPECIFIC_CSS"] = static.statistics_style.read_text(encoding=ENCODING)

    for key, value in variables.items():
        template = template.replace(f"{{{{{key}}}}}", str(value))
    return template



import json

from cognite.neat._utils.http_client import HTTPClient, ParametersRequest
from cognite.neat._utils.http_client._data_classes import SimpleBodyRequest

from .config import NeatClientConfig
from .data_classes import SpaceStatisticsResponse, StatisticsResponse


class StatisticsAPI:
    def __init__(self, neat_config: NeatClientConfig, http_client: HTTPClient) -> None:
        self._config = neat_config
        self._http_client = http_client

    def project(self) -> StatisticsResponse:
        """Retrieve project-wide usage data and limits.

        Returns:
            StatisticsResponse object.
        """

        result = self._http_client.request_with_retries(
            ParametersRequest(
                endpoint_url=self._config.create_api_url("/models/statistics"),
                method="GET",
                parameters=None,
            )
        )

        result.raise_for_status()
        result = StatisticsResponse.model_validate_json(result.success_response.body)
        return result

    def space_statistics(self, spaces: list[str]) -> SpaceStatisticsResponse:
        """Retrieve space-wise usage data and limits.

        Args:
            spaces: List of space identifiers to retrieve statistics for.

        Returns:
            SpaceStatisticsResponse object.
        """

        body = {"items": [{"space": space} for space in spaces]}

        result = self._http_client.request_with_retries(
            SimpleBodyRequest(
                endpoint_url=self._config.create_api_url("/models/statistics/spaces/byids"),
                method="POST",
                body=json.dumps(body),
            )
        )

        result.raise_for_status()
        result = SpaceStatisticsResponse.model_validate_json(result.success_response.body)
        return result



def _is_in_notebook() -> bool:
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def _is_in_browser() -> bool:
    try:
        from pyodide.ffi import IN_BROWSER  # type: ignore [import-not-found]
    except ModuleNotFoundError:
        return False
    return IN_BROWSER


IN_PYODIDE = _is_in_browser()
IN_NOTEBOOK = _is_in_notebook() or IN_PYODIDE



import re
from abc import ABC
from collections.abc import Mapping
from typing import Annotated, Literal

from pydantic import Field, TypeAdapter, field_validator

from cognite.neat._utils.auxiliary import get_concrete_subclasses
from cognite.neat._utils.useful_types import BaseModelObject

from ._constants import ENUM_VALUE_IDENTIFIER_PATTERN, FORBIDDEN_ENUM_VALUES, INSTANCE_ID_PATTERN
from ._references import ContainerReference, ViewReference


class PropertyTypeDefinition(BaseModelObject, ABC):
    type: str


class ListablePropertyTypeDefinition(PropertyTypeDefinition, ABC):
    list: bool | None = Field(
        default=None,
        description="Specifies that the data type is a list of values.",
    )
    max_list_size: int | None = Field(
        default=None,
        description="Specifies the maximum number of values in the list",
    )


class TextProperty(ListablePropertyTypeDefinition):
    type: Literal["text"] = "text"
    max_text_size: int | None = Field(
        default=None,
        description="Specifies the maximum size in bytes of the text property, when encoded with utf-8.",
    )
    collation: str | None = Field(
        default=None,
        description="he set of language specific rules - used when sorting text fields.",
    )


class Unit(BaseModelObject):
    external_id: str = Field(
        description="The external ID of the unit. Must match the unit in the Cognite Unit catalog.",
        min_length=1,
        max_length=256,
        pattern=INSTANCE_ID_PATTERN,
    )
    source_unit: str | None = Field(
        default=None,
        description="The unit in the source system.",
    )


class FloatProperty(ListablePropertyTypeDefinition, ABC):
    unit: Unit | None = Field(default=None, description="The unit of the data stored in this property", exclude=False)


class Float32Property(FloatProperty):
    type: Literal["float32"] = "float32"


class Float64Property(FloatProperty):
    type: Literal["float64"] = "float64"


class BooleanProperty(ListablePropertyTypeDefinition):
    type: Literal["boolean"] = "boolean"


class Int32Property(ListablePropertyTypeDefinition):
    type: Literal["int32"] = "int32"


class Int64Property(ListablePropertyTypeDefinition):
    type: Literal["int64"] = "int64"


class TimestampProperty(ListablePropertyTypeDefinition):
    type: Literal["timestamp"] = "timestamp"


class DateProperty(ListablePropertyTypeDefinition):
    type: Literal["date"] = "date"


class JSONProperty(ListablePropertyTypeDefinition):
    type: Literal["json"] = "json"


class TimeseriesCDFExternalIdReference(ListablePropertyTypeDefinition):
    type: Literal["timeseries"] = "timeseries"


class FileCDFExternalIdReference(ListablePropertyTypeDefinition):
    type: Literal["file"] = "file"


class SequenceCDFExternalIdReference(ListablePropertyTypeDefinition):
    type: Literal["sequence"] = "sequence"


class DirectNodeRelation(ListablePropertyTypeDefinition):
    type: Literal["direct"] = "direct"
    container: ContainerReference | None = Field(
        default=None,
        description="The (optional) required type for the node the direct relation points to.  If specified, "
        "the node must exist before the direct relation is referenced and of the specified type. "
        "If no container specification is used, the node will be auto created with the built-in node "
        "container type, and it does not explicitly have to be created before the node that references it.",
    )
    # This property is only available in the response object. It will be ignored in the request object.
    # In the request object, use ViewCoreProperty.source instead.
    source: ViewReference | None = Field(
        None, description="The hint showing the view what the direct relation points to.", exclude=True
    )


class EnumValue(BaseModelObject):
    name: str | None = Field(
        None,
        max_length=255,
        description="The name of the enum value.",
    )
    description: str | None = Field(
        default=None,
        max_length=1024,
        description="Description of the enum value.",
    )


_ENUM_KEY = re.compile(ENUM_VALUE_IDENTIFIER_PATTERN)


class EnumProperty(PropertyTypeDefinition):
    type: Literal["enum"] = "enum"
    unknown_value: str | None = Field(
        default=None,
        description="TThe value to use when the enum value is unknown. This can optionally be used to "
        "provide forward-compatibility, Specifying what value to use if the client does not "
        "recognize the returned value. It is not possible to ingest the unknown value, "
        "but it must be part of the allowed values.",
        min_length=1,
        max_length=128,
        pattern=ENUM_VALUE_IDENTIFIER_PATTERN,
    )
    values: dict[str, EnumValue] = Field(
        description="A set of all possible values for the enum property.",
        min_length=1,
        max_length=32,
    )

    @field_validator("values", mode="after")
    def _valid_enum_value(cls, val: dict[str, EnumValue]) -> dict[str, EnumValue]:
        errors: list[str] = []
        for key in val.keys():
            if not _ENUM_KEY.match(key):
                errors.append(
                    f"Enum value {key!r} is not valid. Enum values must match "
                    f"the pattern: {ENUM_VALUE_IDENTIFIER_PATTERN}"
                )
            if len(key) > 128 or len(key) < 1:
                errors.append(f"Enum value {key!r} must be between 1 and 128 characters long.")
            if key.lower() in FORBIDDEN_ENUM_VALUES:
                errors.append(
                    f"Enum value {key!r} cannot be any of the following reserved values: {FORBIDDEN_ENUM_VALUES}"
                )
        if errors:
            raise ValueError(";".join(errors))
        return val


DataType = Annotated[
    TextProperty
    | Float32Property
    | Float64Property
    | BooleanProperty
    | Int32Property
    | Int64Property
    | TimestampProperty
    | DateProperty
    | JSONProperty
    | TimeseriesCDFExternalIdReference
    | FileCDFExternalIdReference
    | SequenceCDFExternalIdReference
    | EnumProperty
    | DirectNodeRelation,
    Field(discriminator="type"),
]

DataTypeAdapter: TypeAdapter[DataType] = TypeAdapter(DataType)

DMS_DATA_TYPES: Mapping[str, type[PropertyTypeDefinition]] = {
    cls_.model_fields["type"].default: cls_ for cls_ in get_concrete_subclasses(PropertyTypeDefinition)
}



import sys
from pathlib import Path
from typing import Any, Literal, TypeAlias, get_args

from pydantic import BaseModel, ConfigDict, ValidationError

from cognite.neat._utils.validation import humanize_validation_error

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

LoginFlow: TypeAlias = Literal["client_credentials", "interactive", "token"]
AVAILABLE_LOGIN_FLOWS: tuple[LoginFlow, ...] = get_args(LoginFlow)
Provider: TypeAlias = Literal["entra_id", "auth0", "cdf", "other"]
AVAILABLE_PROVIDERS: tuple[Provider, ...] = get_args(Provider)


class ClientEnvironmentVariables(BaseModel):
    """Configuration for environment variables used by the NEAT client."""

    model_config = ConfigDict(extra="forbid", frozen=True)
    CDF_CLUSTER: str
    CDF_PROJECT: str
    PROVIDER: Provider = "entra_id"
    LOGIN_FLOW: LoginFlow = "client_credentials"

    IDP_CLIENT_ID: str | None = None
    IDP_CLIENT_SECRET: str | None = None
    CDF_TOKEN: str | None = None

    IDP_TENANT_ID: str | None = None
    IDP_TOKEN_URL: str | None = None

    CDF_URL: str | None = None
    IDP_AUDIENCE: str | None = None
    IDP_SCOPES: str | None = None
    IDP_AUTHORITY_URL: str | None = None
    IDP_DISCOVERY_URL: str | None = None
    CDF_MAX_WORKERS: int | None = None
    CDF_CLIENT_TIMEOUT: int | None = None
    CDF_REDIRECT_PORT: int = 53_000

    @classmethod
    def create_humanize(cls, values: dict[str, Any]) -> Self:
        try:
            return cls.model_validate(values)
        except ValidationError as e:
            errors = [humanize_validation_error(error) for error in e.errors()]
            raise ValueError("Invalid environment variable configuration:\n" + "\n - ".join(errors)) from e

    @property
    def idp_tenant_id(self) -> str:
        if self.IDP_TENANT_ID:
            return self.IDP_TENANT_ID
        # This line is technically unreachable due to the checks in idp_token_url and idp_authority_url
        raise RuntimeError("IDP_TENANT_ID is missing")

    @property
    def idp_token_url(self) -> str:
        if self.PROVIDER == "cdf":
            return "https://auth.cognite.com/oauth2/token"
        if self.IDP_TOKEN_URL:
            return self.IDP_TOKEN_URL
        if self.PROVIDER == "entra_id" and self.IDP_TENANT_ID:
            return f"https://login.microsoftonline.com/{self.IDP_TENANT_ID}/oauth2/v2.0/token"
        alternative = " or provide IDP_TENANT_ID" if self.PROVIDER == "entra_id" else ""
        raise ValueError(
            f"IDP_TOKEN_URL is missing. Please provide it{alternative} in the environment variables.",
        )

    @property
    def cdf_url(self) -> str:
        return self.CDF_URL or f"https://{self.CDF_CLUSTER}.cognitedata.com"

    @property
    def idp_audience(self) -> str:
        if self.IDP_AUDIENCE:
            return self.IDP_AUDIENCE
        if self.PROVIDER == "auth0":
            return f"https://{self.CDF_PROJECT}.fusion.cognite.com/{self.CDF_PROJECT}"
        else:
            return f"https://{self.CDF_CLUSTER}.cognitedata.com"

    @property
    def idp_scopes(self) -> list[str]:
        if self.IDP_SCOPES:
            return self.IDP_SCOPES.split(",")
        if self.PROVIDER == "auth0":
            return ["IDENTITY", "user_impersonation"]
        return [f"https://{self.CDF_CLUSTER}.cognitedata.com/.default"]

    @property
    def idp_authority_url(self) -> str:
        if self.IDP_AUTHORITY_URL:
            return self.IDP_AUTHORITY_URL
        if self.PROVIDER == "entra_id" and self.IDP_TENANT_ID:
            return f"https://login.microsoftonline.com/{self.IDP_TENANT_ID}"
        alternative = " or provide IDP_TENANT_ID" if self.PROVIDER == "entra_id" else ""
        raise ValueError(
            f"IDP_AUTHORITY_URL is missing. Please provide it{alternative} in the environment variables.",
        )


def parse_env_file(env_file_path: Path) -> ClientEnvironmentVariables:
    content = env_file_path.read_text()
    variables: dict[str, Any] = {}
    for line in content.splitlines():
        if line.startswith("#") or "=" not in line:
            continue
        key, value = line.strip().split("=", 1)
        variables[key] = value
    return ClientEnvironmentVariables.create_humanize(variables)


def create_env_file_content(provider: Provider, login_flow: LoginFlow) -> str:
    lines = [
        "# Cognite NEAT Client Environment Variables",
        "CDF_CLUSTER=<your-cdf-cluster>",
        "CDF_PROJECT=<your-cdf-project>",
        "",
    ]
    if login_flow != "token":
        lines.append(f"PROVIDER={provider}")
    lines.append(f"LOGIN_FLOW={login_flow}")
    lines.append("")
    if login_flow in ("client_credentials", "interactive"):
        lines.append("IDP_CLIENT_ID=<your-idp-client-id>")
        if login_flow == "client_credentials":
            lines.append("IDP_CLIENT_SECRET=<your-idp-client-secret>")
        if provider == "entra_id":
            lines.append("IDP_TENANT_ID=<your-idp-tenant-id>")
        if provider not in ("cdf", "entra_id"):
            lines.append("IDP_TOKEN_URL=<your-idp-token-url>")
        if provider == "other":
            lines.append("IDP_AUDIENCE=<your-idp-audience>")
            lines.append("IDP_SCOPES=<your-idp-scopes-comma-separated>")
            lines.append("IDP_AUTHORITY_URL=<your-idp-authority-url>")
    elif login_flow == "token":
        lines.append("CDF_TOKEN=<your-cdf-token>")

    return "\n".join(lines)



from cognite.neat._data_model.deployer.data_classes import DeploymentResult
from cognite.neat._session._html._render import render
from cognite.neat._store import NeatStore

from ._deployment._physical.serializer import serialize_deployment_result


class Result:
    """Class to handle deployment results in the NeatSession."""

    def __init__(self, store: NeatStore) -> None:
        self._store = store

    @property
    def _result(self) -> DeploymentResult | None:
        """Get deployment result from the last change in the store."""
        if change := self._store.provenance.last_change:
            if change.result:
                return change.result
        return None

    def _repr_html_(self) -> str:
        """Generate interactive HTML representation."""
        if not self._result:
            return "<p>No deployment result available</p>"

        if isinstance(self._result, DeploymentResult):
            serialized_result = serialize_deployment_result(self._result)
            return render("deployment", serialized_result)
        else:
            raise NotImplementedError(f"HTML rendering for the result type {type(self._result)} is not implemented.")



import itertools
import warnings
from collections import defaultdict
from collections.abc import Hashable, ItemsView, Iterator, KeysView, MutableMapping, Set, ValuesView
from dataclasses import dataclass, field
from graphlib import TopologicalSorter
from typing import Any, Literal, TypeVar, overload

import networkx as nx
import pandas as pd
from cognite.client import data_modeling as dm
from rdflib import URIRef

from cognite.neat._v0.core._data_model.models import ConceptualDataModel, PhysicalDataModel
from cognite.neat._v0.core._data_model.models.conceptual import (
    Concept,
    ConceptualProperty,
)
from cognite.neat._v0.core._data_model.models.entities import (
    ConceptEntity,
    MultiValueTypeInfo,
    ViewEntity,
)
from cognite.neat._v0.core._data_model.models.entities._single_value import (
    UnknownEntity,
)
from cognite.neat._v0.core._data_model.models.physical import PhysicalProperty
from cognite.neat._v0.core._data_model.models.physical._verified import PhysicalView
from cognite.neat._v0.core._issues.errors import NeatValueError
from cognite.neat._v0.core._issues.warnings import NeatValueWarning

T_Hashable = TypeVar("T_Hashable", bound=Hashable)


@dataclass(frozen=True)
class Linkage:
    source_concept: ConceptEntity
    connecting_property: str
    target_concept: ConceptEntity
    max_occurrence: int | float | None


class LinkageSet(set, Set[Linkage]):
    @property
    def source_concept(self) -> set[ConceptEntity]:
        return {link.source_concept for link in self}

    @property
    def target_concept(self) -> set[ConceptEntity]:
        return {link.target_concept for link in self}

    def get_target_concepts_by_source(self) -> dict[ConceptEntity, set[ConceptEntity]]:
        target_concepts_by_source: dict[ConceptEntity, set[ConceptEntity]] = defaultdict(set)
        for link in self:
            target_concepts_by_source[link.source_concept].add(link.target_concept)
        return target_concepts_by_source

    def to_pandas(self) -> pd.DataFrame:
        # Todo: Remove this method
        return pd.DataFrame(
            [
                {
                    "source_concept": link.source_concept,
                    "connecting_property": link.connecting_property,
                    "target_concept": link.target_concept,
                    "max_occurrence": link.max_occurrence,
                }
                for link in self
            ]
        )


@dataclass
class ViewQuery:
    view_id: dm.ViewId
    rdf_type: URIRef
    property_renaming_config: dict[URIRef, str] = field(default_factory=dict)


class ViewQueryDict(dict, MutableMapping[dm.ViewId, ViewQuery]):
    # The below methods are included to make better type hints in the IDE
    def __getitem__(self, k: dm.ViewId) -> ViewQuery:
        return super().__getitem__(k)

    def __setitem__(self, k: dm.ViewId, v: ViewQuery) -> None:
        super().__setitem__(k, v)

    def __delitem__(self, k: dm.ViewId) -> None:
        super().__delitem__(k)

    def __iter__(self) -> Iterator[dm.ViewId]:
        return super().__iter__()

    def keys(self) -> KeysView[dm.ViewId]:  # type: ignore[override]
        return super().keys()

    def values(self) -> ValuesView[ViewQuery]:  # type: ignore[override]
        return super().values()

    def items(self) -> ItemsView[dm.ViewId, ViewQuery]:  # type: ignore[override]
        return super().items()

    def get(self, __key: dm.ViewId, __default: Any = ...) -> ViewQuery:
        return super().get(__key, __default)

    def pop(self, __key: dm.ViewId, __default: Any = ...) -> ViewQuery:
        return super().pop(__key, __default)

    def popitem(self) -> tuple[dm.ViewId, ViewQuery]:
        return super().popitem()


class DataModelAnalysis:
    def __init__(
        self,
        conceptual: ConceptualDataModel | None = None,
        physical: PhysicalDataModel | None = None,
    ) -> None:
        self._conceptual = conceptual
        self._physical = physical

    @property
    def conceptual(self) -> ConceptualDataModel:
        if self._conceptual is None:
            raise NeatValueError("Conceptual Data Model is required for this analysis")
        return self._conceptual

    @property
    def physical(self) -> PhysicalDataModel:
        if self._physical is None:
            raise NeatValueError("Physical Data Model is required for this analysis")
        return self._physical

    def parents_by_concept(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ConceptEntity, set[ConceptEntity]]:
        """Get a dictionary of concepts and their parents.

        Args:
            include_ancestors (bool, optional): Include ancestors of the parents. Defaults to False.
            include_different_space (bool, optional): Include parents from different spaces. Defaults to False.

        Returns:
            dict[ConceptEntity, set[ConceptEntity]]: Values parents with concept as key.
        """
        parents_by_concept: dict[ConceptEntity, set[ConceptEntity]] = {}
        for concept in self.conceptual.concepts:
            parents_by_concept[concept.concept] = set()
            for parent in concept.implements or []:
                if include_different_space or parent.prefix == concept.concept.prefix:
                    parents_by_concept[concept.concept].add(parent)
                else:
                    warnings.warn(
                        NeatValueWarning(
                            f"Parent concept {parent} of concept {concept} is not in the same namespace, skipping!"
                        ),
                        stacklevel=2,
                    )
        if include_ancestors:
            self._include_ancestors(parents_by_concept)

        return parents_by_concept

    @staticmethod
    def _include_ancestors(
        parents_by_concept: dict[T_Hashable, set[T_Hashable]],
    ) -> None:
        # Topological sort to ensure that concepts include all ancestors
        for concept_entity in list(TopologicalSorter(parents_by_concept).static_order()):
            if concept_entity not in parents_by_concept:
                continue
            parents_by_concept[concept_entity] |= {
                grand_parent
                for parent in parents_by_concept[concept_entity]
                for grand_parent in parents_by_concept.get(parent, set())
            }

    def properties_by_concepts(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ConceptEntity, list[ConceptualProperty]]:
        """Get a dictionary of concepts and their properties.

        Args:
            include_ancestors: Whether to include properties from parent concepts.
            include_different_space: Whether to include properties from parent concepts in different spaces.

        Returns:
            dict[ConceptEntity, list[ConceptualProperty]]: Values properties with concept as key.

        """
        properties_by_concepts: dict[ConceptEntity, list[ConceptualProperty]] = defaultdict(list)
        for prop in self.conceptual.properties:
            properties_by_concepts[prop.concept].append(prop)

        if include_ancestors:
            parents_by_concepts = self.parents_by_concept(
                include_ancestors=include_ancestors,
                include_different_space=include_different_space,
            )
            for concept, parents in parents_by_concepts.items():
                concept_properties = {prop.property_ for prop in properties_by_concepts[concept]}
                for parent in parents:
                    for parent_prop in properties_by_concepts[parent]:
                        if parent_prop.property_ not in concept_properties:
                            child_prop = parent_prop.model_copy(update={"concept": concept})
                            properties_by_concepts[concept].append(child_prop)
                            concept_properties.add(child_prop.property_)

        return properties_by_concepts

    def implements_by_view(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ViewEntity, set[ViewEntity]]:
        """Get a dictionary of views and their implemented views."""
        # This is a duplicate fo the parent_by_concept method, but for views
        # The choice to duplicate the code is to avoid generics which will make the code less readable
        implements_by_view: dict[ViewEntity, set[ViewEntity]] = {}
        for view in self.physical.views:
            implements_by_view[view.view] = set()
            for implements in view.implements or []:
                if include_different_space or implements.space == view.view.space:
                    implements_by_view[view.view].add(implements)
                else:
                    warnings.warn(
                        NeatValueWarning(
                            f"Implemented view {implements} of view {view} is not in the same namespace, skipping!"
                        ),
                        stacklevel=2,
                    )
        if include_ancestors:
            self._include_ancestors(implements_by_view)
        return implements_by_view

    def properties_by_view(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ViewEntity, list[PhysicalProperty]]:
        """Get a dictionary of views and their properties."""
        # This is a duplicate fo the properties_by_concept method, but for views
        # The choice to duplicate the code is to avoid generics which will make the code less readable.
        properties_by_views: dict[ViewEntity, list[PhysicalProperty]] = defaultdict(list)
        for prop in self.physical.properties:
            properties_by_views[prop.view].append(prop)

        if include_ancestors:
            implements_by_view = self.implements_by_view(
                include_ancestors=include_ancestors, include_different_space=include_different_space
            )
            for view, parents in implements_by_view.items():
                view_properties = {prop.view_property for prop in properties_by_views[view]}
                for parent in parents:
                    for parent_prop in properties_by_views[parent]:
                        if parent_prop.view_property not in view_properties:
                            child_prop = parent_prop.model_copy(update={"view": view})
                            properties_by_views[view].append(child_prop)
                            view_properties.add(child_prop.view_property)

        return properties_by_views

    @property
    def conceptual_uri_by_view(self) -> dict[ViewEntity, URIRef]:
        """Get the logical URI by view."""
        return {view.view: view.conceptual for view in self.physical.views if view.conceptual}

    def conceptual_uri_by_property_by_view(
        self,
        include_ancestors: bool = False,
        include_different_space: bool = False,
    ) -> dict[ViewEntity, dict[str, URIRef]]:
        """Get the logical URI by property by view."""
        properties_by_view = self.properties_by_view(include_ancestors, include_different_space)

        return {
            view: {prop.view_property: prop.conceptual for prop in properties if prop.conceptual}
            for view, properties in properties_by_view.items()
        }

    @property
    def _concept_by_neat_id(self) -> dict[URIRef, Concept]:
        """Get a dictionary of concept neat IDs to
        concept entities."""

        return {cls.neatId: cls for cls in self.conceptual.concepts if cls.neatId}

    def concept_by_suffix(self) -> dict[str, Concept]:
        """Get a dictionary of concept suffixes to concept entities."""
        # TODO: Remove this method
        concept_dict: dict[str, Concept] = {}
        for definition in self.conceptual.concepts:
            entity = definition.concept
            if entity.suffix in concept_dict:
                warnings.warn(
                    NeatValueWarning(
                        f"Concept {entity} has been defined more than once! Only the first definition "
                        "will be considered, skipping the rest.."
                    ),
                    stacklevel=2,
                )
                continue
            concept_dict[entity.suffix] = definition
        return concept_dict

    @property
    def concept_by_concept_entity(self) -> dict[ConceptEntity, Concept]:
        """Get a dictionary of concept entities to concept entities."""
        data_model = self.conceptual
        return {cls.concept: cls for cls in data_model.concepts}

    @property
    def view_by_view_entity(self) -> dict[ViewEntity, PhysicalView]:
        """Get a dictionary of views to view entities."""
        data_model = self.physical
        return {view.view: view for view in data_model.views}

    def property_by_id(self) -> dict[str, list[ConceptualProperty]]:
        """Get a dictionary of property IDs to property entities."""
        property_dict: dict[str, list[ConceptualProperty]] = defaultdict(list)
        for prop in self.conceptual.properties:
            property_dict[prop.property_].append(prop)
        return property_dict

    def properties_by_id_by_concept(
        self,
        has_instance_source: bool = False,
        include_ancestors: bool = False,
    ) -> dict[ConceptEntity, dict[str, ConceptualProperty]]:
        """Get a dictionary of concept entities to dictionaries of property IDs to property entities."""
        concept_property_pairs: dict[ConceptEntity, dict[str, ConceptualProperty]] = {}
        for concept, properties in self.properties_by_concepts(include_ancestors).items():
            processed_properties: dict[str, ConceptualProperty] = {}
            for prop in properties:
                if prop.property_ in processed_properties:
                    warnings.warn(
                        NeatValueWarning(
                            f"Property {processed_properties} for {concept} has been defined more than once!"
                            " Only the first definition will be considered, skipping the rest.."
                        ),
                        stacklevel=2,
                    )
                    continue
                if has_instance_source and prop.instance_source is None:
                    continue
                processed_properties[prop.property_] = prop
            concept_property_pairs[concept] = processed_properties

        return concept_property_pairs

    def defined_views(self, include_ancestors: bool = False) -> set[ViewEntity]:
        properties_by_view = self.properties_by_view(include_ancestors)
        return {prop.view for prop in itertools.chain.from_iterable(properties_by_view.values())}

    def concepts(self) -> set[ConceptEntity]:
        """Returns all concepts defined Concepts in the data model."""
        return {concept.concept for concept in self.conceptual.concepts}

    def defined_concepts(
        self,
        include_ancestors: bool = False,
    ) -> set[ConceptEntity]:
        """Returns concepts that have properties defined for them in the data model.

        Args:
            include_ancestors: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of concepts that have been defined in the data model
        """
        properties_by_concept = self.properties_by_concepts(include_ancestors)
        return {prop.concept for prop in itertools.chain.from_iterable(properties_by_concept.values())}

    def concept_linkage(
        self,
        include_ancestors: bool = False,
    ) -> LinkageSet:
        """Returns a set of concept linkages in the data model.

        Args:
            include_ancestors: Whether to consider inheritance or not. Defaults False

        Returns:

        """
        concept_linkage = LinkageSet()

        properties_by_concept = self.properties_by_concepts(include_ancestors)

        prop: ConceptualProperty
        for prop in itertools.chain.from_iterable(properties_by_concept.values()):
            if not isinstance(prop.value_type, ConceptEntity):
                continue
            concept_linkage.add(
                Linkage(
                    source_concept=prop.concept,
                    connecting_property=prop.property_,
                    target_concept=prop.value_type,
                    max_occurrence=prop.max_count,
                )
            )

        return concept_linkage

    def symmetrically_connected_concepts(
        self,
        include_ancestors: bool = False,
    ) -> set[tuple[ConceptEntity, ConceptEntity]]:
        """Returns a set of pairs of symmetrically linked concepts.

        Args:
            include_ancestors: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of pairs of symmetrically linked concepts

        !!! note "Symmetrically Connected Concepts"
            Symmetrically connected concepts are concepts that are connected to each other
            in both directions. For example, if concept A is connected to concept B, and concept B
            is connected to concept A, then concepts A and B are symmetrically connected.
        """
        sym_pairs: set[tuple[ConceptEntity, ConceptEntity]] = set()
        concept_linkage = self.concept_linkage(include_ancestors)
        if not concept_linkage:
            return sym_pairs

        targets_by_source = concept_linkage.get_target_concepts_by_source()
        for link in concept_linkage:
            source = link.source_concept
            target = link.target_concept

            if source in targets_by_source[source] and (source, target) not in sym_pairs:
                sym_pairs.add((source, target))
        return sym_pairs

    @overload
    def _properties_by_neat_id(self, format: Literal["info"] = "info") -> dict[URIRef, ConceptualProperty]: ...

    @overload
    def _properties_by_neat_id(self, format: Literal["dms"] = "dms") -> dict[URIRef, PhysicalProperty]: ...

    def _properties_by_neat_id(
        self, format: Literal["info", "dms"] = "info"
    ) -> dict[URIRef, ConceptualProperty] | dict[URIRef, PhysicalProperty]:
        if format == "info":
            return {prop.neatId: prop for prop in self.conceptual.properties if prop.neatId}
        elif format == "dms":
            return {prop.neatId: prop for prop in self.physical.properties if prop.neatId}
        else:
            raise NeatValueError(f"Invalid format: {format}")

    @property
    def concepts_by_neat_id(self) -> dict[URIRef, Concept]:
        return {concept.neatId: concept for concept in self.conceptual.concepts if concept.neatId}

    @property
    def multi_value_properties(self) -> list[ConceptualProperty]:
        return [prop_ for prop_ in self.conceptual.properties if isinstance(prop_.value_type, MultiValueTypeInfo)]

    @property
    def view_query_by_id(
        self,
    ) -> "ViewQueryDict":
        # Trigger error if any of these are missing
        _ = self.conceptual
        _ = self.physical

        # caching results for faster access
        concepts_by_neat_id = self._concept_by_neat_id
        properties_by_concept = self.properties_by_concepts(include_ancestors=True)
        conceptual_uri_by_view = self.conceptual_uri_by_view
        conceptual_uri_by_property_by_view = self.conceptual_uri_by_property_by_view(include_ancestors=True)
        conceptual_properties_by_neat_id = self._properties_by_neat_id()

        query_configs = ViewQueryDict()
        for view in self.physical.views:
            # this entire block of sequential if statements checks:
            # 1. connection of physical and conceptual data model
            # 2. correct paring of conceptual and physical data model
            # 3. connection of conceptual data model to instances
            if (
                (neat_id := conceptual_uri_by_view.get(view.view))
                and (concept := concepts_by_neat_id.get(neat_id))
                and (uri := concept.instance_source)
            ):
                view_query = ViewQuery(
                    view_id=view.view.as_id(),
                    rdf_type=uri,
                    # start off with renaming of properties on the information level
                    # this is to encounter for special cases of e.g. space, startNode and endNode
                    property_renaming_config=(
                        {uri: prop_.property_ for prop_ in info_properties for uri in prop_.instance_source or []}
                        if (info_properties := properties_by_concept.get(concept.concept))
                        else {}
                    ),
                )

                if conceptual_uri_by_property := conceptual_uri_by_property_by_view.get(view.view):
                    for target_name, neat_id in conceptual_uri_by_property.items():
                        if (property_ := conceptual_properties_by_neat_id.get(neat_id)) and (
                            uris := property_.instance_source
                        ):
                            for uri in uris:
                                view_query.property_renaming_config[uri] = target_name

                query_configs[view.view.as_id()] = view_query

        return query_configs

    def _physical_di_graph(self, format: Literal["data-model", "implements"] = "data-model") -> nx.MultiDiGraph:
        """Generate a MultiDiGraph from the Physical Data Model."""
        di_graph = nx.MultiDiGraph()

        data_model = self.physical

        # Add nodes and edges from Views sheet
        for view in data_model.views:
            di_graph.add_node(view.view.suffix, label=view.view.suffix)

            if format == "implements" and view.implements:
                for implement in view.implements:
                    di_graph.add_node(implement.suffix, label=implement.suffix)
                    di_graph.add_edge(
                        view.view.suffix,
                        implement.suffix,
                        label="implements",
                        dashes=True,
                    )

        if format == "data-model":
            # Add nodes and edges from Properties sheet
            for prop_ in data_model.properties:
                if prop_.connection and isinstance(prop_.value_type, ViewEntity):
                    di_graph.add_node(prop_.view.suffix, label=prop_.view.suffix)
                    di_graph.add_node(prop_.value_type.suffix, label=prop_.value_type.suffix)
                    di_graph.add_edge(
                        prop_.view.suffix,
                        prop_.value_type.suffix,
                        label=prop_.name or prop_.view_property,
                    )

        return di_graph

    def _conceptual_di_graph(self, format: Literal["data-model", "implements"] = "data-model") -> nx.MultiDiGraph:
        """Generate MultiDiGraph representing conceptual data model."""

        data_model = self.conceptual
        di_graph = nx.MultiDiGraph()

        # Add nodes and edges from Views sheet
        for concept in data_model.concepts:
            # if possible use human readable label coming from the view name

            di_graph.add_node(
                concept.concept.suffix,
                label=concept.name or concept.concept.suffix,
            )

            if format == "implements" and concept.implements:
                for parent in concept.implements:
                    di_graph.add_node(parent.suffix, label=parent.suffix)
                    di_graph.add_edge(
                        concept.concept.suffix,
                        parent.suffix,
                        label="implements",
                        dashes=True,
                    )

        if format == "data-model":
            # Add nodes and edges from Properties sheet
            for prop_ in data_model.properties:
                if isinstance(prop_.value_type, ConceptEntity) and not isinstance(prop_.value_type, UnknownEntity):
                    di_graph.add_node(prop_.concept.suffix, label=prop_.concept.suffix)
                    di_graph.add_node(prop_.value_type.suffix, label=prop_.value_type.suffix)

                    di_graph.add_edge(
                        prop_.concept.suffix,
                        prop_.value_type.suffix,
                        label=prop_.name or prop_.property_,
                    )

        return di_graph



from ._api_exporter import DMSAPIExporter, DMSAPIJSONExporter, DMSAPIYAMLExporter
from ._base import DMSExporter, DMSFileExporter
from ._table_exporter.exporter import DMSExcelExporter, DMSTableExporter, DMSTableJSONExporter, DMSTableYamlExporter

__all__ = [
    "DMSAPIExporter",
    "DMSAPIJSONExporter",
    "DMSAPIYAMLExporter",
    "DMSExcelExporter",
    "DMSExporter",
    "DMSFileExporter",
    "DMSTableExporter",
    "DMSTableJSONExporter",
    "DMSTableYamlExporter",
]



from collections.abc import Sequence

from cognite.neat._data_model.models.dms import DataModelReference, DataModelRequest, DataModelResponse
from cognite.neat._utils.http_client import HTTPClient, SuccessResponse

from .api import Endpoint, NeatAPI
from .config import NeatClientConfig
from .data_classes import PagedResponse
from .filters import DataModelFilter


class DataModelsAPI(NeatAPI):
    def __init__(self, neat_config: NeatClientConfig, http_client: HTTPClient) -> None:
        super().__init__(
            neat_config,
            http_client,
            endpoint_map={
                "apply": Endpoint("POST", "/models/datamodels", item_limit=100),
                "retrieve": Endpoint("POST", "/models/datamodels/byids", item_limit=100),
                "delete": Endpoint("POST", "/models/datamodels/delete", item_limit=100),
                "list": Endpoint("GET", "/models/datamodels", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse) -> PagedResponse[DataModelResponse]:
        return PagedResponse[DataModelResponse].model_validate_json(response.body)

    def _validate_id_response(self, response: SuccessResponse) -> list[DataModelReference]:
        return PagedResponse[DataModelReference].model_validate_json(response.body).items

    def apply(self, data_models: Sequence[DataModelRequest]) -> list[DataModelResponse]:
        """Apply (create or update) data models in CDF.

        Args:
            data_models: List of DataModelRequest objects to apply.
        Returns:
            List of DataModelResponse objects.
        """
        return self._request_item_response(data_models, "apply")

    def retrieve(self, items: list[DataModelReference]) -> list[DataModelResponse]:
        """Retrieve data models by their identifiers.

        Args:
            items: List of data models references identifying the data models to retrieve.
        Returns:
            List of DataModelResponse objects.
        """
        return self._request_item_response(items, "retrieve")

    def delete(self, items: list[DataModelReference]) -> list[DataModelReference]:
        """Delete data models by their identifiers.

        Args:
            items: List of data model references identifying the data models to delete.
        Returns:
            List of DataModelReference objects representing the deleted data models.
        """
        return self._request_id_response(items, "delete")

    def list(
        self,
        space: str | None = None,
        all_versions: bool = False,
        inline_views: bool = False,
        include_global: bool = False,
        limit: int | None = 10,
    ) -> list[DataModelResponse]:
        """List data models in CDF Project.

        Args:
            space: If specified, only data models in this space are returned.
            all_versions: If True, return all versions. If False, only return the latest version.
            inline_views: If True, include views inline in the response.
            include_global: If True, include global data models.
            limit: Maximum number of data models to return. If None, return all data models.

        Returns:
            List of DataModelResponse objects.
        """
        filter = DataModelFilter(
            space=space,
            all_versions=all_versions,
            inline_views=inline_views,
            include_global=include_global,
        )
        return self._list(limit=limit, params=filter.dump())
